@@ -86,6 +86,8 @@ bool external_log_open = false;
 int active_control_conns = 0;
 int socket_ready_fd = -1;
 
+sigset_t orig_signal_mask; // signal mask when started
+
 // Control socket path. We maintain a string (control_socket_str) in case we need
 // to allocate storage, but control_socket_path is the authoritative value.
 static const char *control_socket_path = SYSCONTROLSOCKET;
@@ -224,10 +226,39 @@ static int process_commandline_arg(char **argv, int argc, int &i, options &opts)
     service_dir_opt &service_dir_opts = opts.service_dir_opts;
     list<const char *> &services_to_start = opts.services_to_start;
 
+    auto arg_to_loglevel = [&](const char *option_name, loglevel_t &wanted_level) {
+        if (++i < argc && argv[i][0] != '\0') {
+            if (strcmp(argv[i], "none") == 0) {
+                wanted_level = loglevel_t::ZERO;
+            }
+            else if (strcmp(argv[i], "error") == 0) {
+                wanted_level = loglevel_t::ERROR;
+            }
+            else if (strcmp(argv[i], "warn") == 0) {
+                wanted_level = loglevel_t::WARN;
+            }
+            else if (strcmp(argv[i], "info") == 0) {
+                wanted_level = loglevel_t::NOTICE;
+            }
+            else if (strcmp(argv[i], "debug") == 0) {
+                wanted_level = loglevel_t::DEBUG;
+            }
+            else {
+                cerr << "dinit: '" << option_name << "' accepts only arguments: 'none', 'error', 'warn', 'info', 'debug'\n";
+                return false;
+            }
+            return true;
+        }
+        else {
+            cerr << "dinit: '" << option_name << "' requires an argument\n";
+            return false;
+        }
+    };
+
     if (argv[i][0] == '-') {
         // An option...
         if (strcmp(argv[i], "--env-file") == 0 || strcmp(argv[i], "-e") == 0) {
-            if (++i < argc) {
+            if (++i < argc && argv[i][0] != '\0') {
                 env_file_set = true;
                 env_file = argv[i];
             }
@@ -237,7 +268,7 @@ static int process_commandline_arg(char **argv, int argc, int &i, options &opts)
             }
         }
         else if (strcmp(argv[i], "--services-dir") == 0 || strcmp(argv[i], "-d") == 0) {
-            if (++i < argc) {
+            if (++i < argc && argv[i][0] != '\0') {
                 service_dir_opts.set_specified_service_dir(argv[i]);
             }
             else {
@@ -260,7 +291,7 @@ static int process_commandline_arg(char **argv, int argc, int &i, options &opts)
             opts.process_sys_args = false;
         }
         else if (strcmp(argv[i], "--socket-path") == 0 || strcmp(argv[i], "-p") == 0) {
-            if (++i < argc) {
+            if (++i < argc && argv[i][0] != '\0') {
                 control_socket_path = argv[i];
                 control_socket_path_set = true;
             }
@@ -296,7 +327,7 @@ static int process_commandline_arg(char **argv, int argc, int &i, options &opts)
             }
         }
         else if (strcmp(argv[i], "--log-file") == 0 || strcmp(argv[i], "-l") == 0) {
-            if (++i < argc) {
+            if (++i < argc && argv[i][0] != '\0') {
                 log_path = argv[i];
                 log_is_syslog = false;
                 log_specified = true;
@@ -310,9 +341,19 @@ static int process_commandline_arg(char **argv, int argc, int &i, options &opts)
             console_service_status = false;
             log_level[DLOG_CONS] = loglevel_t::ZERO;
         }
+        else if (strcmp(argv[i], "--console-level") == 0) {
+            loglevel_t wanted_level;
+            if (!arg_to_loglevel("--console-level", wanted_level)) return 1;
+            log_level[DLOG_CONS] = wanted_level;
+        }
+        else if (strcmp(argv[i], "--log-level") == 0) {
+            loglevel_t wanted_level;
+            if (!arg_to_loglevel("--log-level", wanted_level)) return 1;
+            log_level[DLOG_MAIN] = wanted_level;
+        }
         #ifdef SUPPORT_CGROUPS
         else if (strcmp(argv[i], "--cgroup-path") == 0 || strcmp(argv[i], "-b") == 0) {
-            if (++i < argc) {
+            if (++i < argc && argv[i][0] != '\0') {
                 cgroups_path = argv[i];
                 have_cgroups_path = true;
             }
@@ -323,7 +364,7 @@ static int process_commandline_arg(char **argv, int argc, int &i, options &opts)
         }
         #endif
         else if (strcmp(argv[i], "--service") == 0 || strcmp(argv[i], "-t") == 0) {
-            if (++i < argc) {
+            if (++i < argc && argv[i][0] != '\0') {
                 services_to_start.push_back(argv[i]);
             }
             else {
@@ -371,6 +412,10 @@ static int process_commandline_arg(char **argv, int argc, int &i, options &opts)
         }
     }
     else {
+        if (argv[i][0] == '\0') {
+            cerr << "dinit: error: empty command-line argument\n";
+            return 1;
+        }
 #ifdef __linux__
         // If we are running as init (PID=1), the Linux kernel gives us all command line arguments it was
         // given but didn't recognize, and, uh, *some* that it did recognize, which means we can't assume
@@ -466,12 +511,18 @@ int dinit_main(int argc, char **argv)
 
     /* Set up signal handlers etc */
     sigset_t sigwait_set;
-    sigemptyset(&sigwait_set);
-    sigaddset(&sigwait_set, SIGCHLD);
-    sigaddset(&sigwait_set, SIGINT);
-    sigaddset(&sigwait_set, SIGTERM);
-    if (am_system_mgr) sigaddset(&sigwait_set, SIGQUIT);
-    sigprocmask(SIG_BLOCK, &sigwait_set, NULL);
+    if (am_system_mgr) {
+        // Block all signals in system manager mode - don't want to chance provoking a signal that
+        // will suspend or terminate the process
+        sigfillset(&sigwait_set);
+    }
+    else {
+        sigemptyset(&sigwait_set);
+        sigaddset(&sigwait_set, SIGCHLD);
+        sigaddset(&sigwait_set, SIGINT);
+        sigaddset(&sigwait_set, SIGTERM);
+    }
+    sigprocmask(SIG_BLOCK, &sigwait_set, &orig_signal_mask);
 
     // Terminal access control signals - we ignore these so that dinit can't be
     // suspended if it writes to the terminal after some other process has claimed
@@ -574,12 +625,17 @@ int dinit_main(int argc, char **argv)
         log(loglevel_t::NOTICE, false, "Starting system");
     }
     
-    // Only try to set up the external log now if we aren't the system init. (If we are the
-    // system init, wait until the log service starts).
-    if (!am_system_init && log_specified) setup_external_log();
+    // If a log file was specified, open it now.
+    if (log_specified) {
+        setup_external_log();
+        if (!am_system_init && !external_log_open) {
+            flush_log(); // flush console messages
+            return EXIT_FAILURE;
+        }
+    }
 
     if (env_file != nullptr) {
-        read_env_file(env_file, true, main_env);
+        read_env_file(env_file, true, main_env, false);
     }
 
     for (auto svc : services_to_start) {
@@ -597,7 +653,7 @@ int dinit_main(int argc, char **argv)
             log_service_load_failure(sde);
         }
         catch (service_load_exc &sle) {
-            log(loglevel_t::ERROR, sle.service_name, ": ", sle.exc_description);
+            log(loglevel_t::ERROR, sle.service_name, ": error loading: ", sle.exc_description);
         }
         catch (std::bad_alloc &badalloce) {
             log(loglevel_t::ERROR, "Out of memory when trying to start service: ", svc, ".");
@@ -693,7 +749,7 @@ int dinit_main(int argc, char **argv)
         sigprocmask(SIG_UNBLOCK, &sigwait_set_int, NULL);
     }
     
-    return 0;
+    return EXIT_SUCCESS;
 }
 
 // Get user confirmation before proceeding with restarting boot sequence.
@@ -786,7 +842,8 @@ static void control_socket_cb(eventloop_t *loop, int sockfd) noexcept
     }
 }
 
-static void control_socket_ready() noexcept {
+static void control_socket_ready() noexcept
+{
     if (!control_socket_open || socket_ready_fd < 0) {
         return;
     }
@@ -795,7 +852,7 @@ static void control_socket_ready() noexcept {
     if (socket_ready_fd > 2) {
         close(socket_ready_fd);
     }
-    // Ensure this can only be called once
+    // Ensure that we don't try to issue readiness again:
     socket_ready_fd = -1;
 }
 
@@ -803,15 +860,17 @@ static void control_socket_ready() noexcept {
 void rootfs_is_rw() noexcept
 {
     open_control_socket(true);
-    if (! did_log_boot) {
+    control_socket_ready();
+    if (!log_is_syslog && !external_log_open) {
+        // Try (again) to open log file if we couldn't do so earlier.
+        setup_external_log();
+    }
+    if (!did_log_boot) {
         did_log_boot = log_boot();
     }
-    // If the control socket failed to open early on, there was no readiness
-    // notification, so do it here for a second time, just in case
-    control_socket_ready();
 }
 
-// Open/create the control socket, normally /dev/dinitctl, used to allow client programs to connect
+// Open/create the control socket, normally /run/dinitctl, used to allow client programs to connect
 // and issue service orders and shutdown commands etc. This can safely be called multiple times;
 // once the socket has been successfully opened, further calls will check the socket file is still
 // present and re-create it if not.
@@ -930,7 +989,7 @@ static void close_control_socket() noexcept
 
 void setup_external_log() noexcept
 {
-    if (! external_log_open) {
+    if (!external_log_open) {
         if (log_is_syslog) {
             const char * saddrname = log_path;
             size_t saddrname_len = strlen(saddrname);
@@ -989,7 +1048,7 @@ void setup_external_log() noexcept
             else {
                 // log failure to log? It makes more sense than first appears, because we also log
                 // to console:
-                log(loglevel_t::ERROR, "Setting up log failed: ", strerror(errno));
+                log(loglevel_t::ERROR, "Opening log file failed: ", strerror(errno));
             }
         }
     }
@@ -1109,6 +1168,9 @@ static void printVersion()
 #endif
 #ifdef USE_UTMPX
             +1
+#endif
+#if USE_INITGROUPS
+            +1
 #endif
             ;
     if (feature_count != 0) {
@@ -1118,6 +1180,9 @@ static void printVersion()
 #endif
 #ifdef USE_UTMPX
                 " utmp"
+#endif
+#if USE_INITGROUPS
+                " supplemental-groups"
 #endif
                 "\n";
     }
