@@ -5,9 +5,9 @@
 #include <sys/types.h>
 #include <sys/resource.h>
 
-#include "baseproc-sys.h"
-#include "service.h"
-#include "dinit-utmp.h"
+#include <baseproc-sys.h>
+#include <service.h>
+#include <dinit-utmp.h>
 
 // This header defines base_proc_service (base process service) and several derivatives, as well as some
 // utility functions and classes. See service.h for full details of services.
@@ -29,25 +29,28 @@ struct run_proc_params
     const char *logfile;      // log file or nullptr (stdout/stderr); must be valid if !on_console
     const char *env_file;     // file with environment settings (or nullptr)
     #if SUPPORT_CGROUPS
-    const char *run_in_cgroup; //  cgroup path
+    const char *run_in_cgroup = nullptr; //  cgroup path
     #endif
     bool on_console;          // whether to run on console
     bool in_foreground;       // if on console: whether to run in foreground
+    bool unmask_sigint = false; // if in foreground: whether to unmask SIGINT
     int wpipefd;              // pipe to which error status will be sent (if error occurs)
     int csfd;                 // control socket fd (or -1); may be moved
     int socket_fd;            // pre-opened socket fd (or -1); may be moved
     int notify_fd;            // pipe for readiness notification message (or -1); may be moved
     int force_notify_fd;      // if not -1, notification fd must be moved to this fd
+    int output_fd;            // if not -1, output will be directed here (rather than logfile)
     const char *notify_var;   // environment variable name where notification fd will be stored, or nullptr
     uid_t uid;
     gid_t gid;
     const std::vector<service_rlimits> &rlimits;
+    int input_fd = -1;        // file descriptor to be used for input (STDIN)
 
     run_proc_params(const char * const *args, const char *working_dir, const char *logfile, int wpipefd,
             uid_t uid, gid_t gid, const std::vector<service_rlimits> &rlimits)
             : args(args), working_dir(working_dir), logfile(logfile), env_file(nullptr), on_console(false),
               in_foreground(false), wpipefd(wpipefd), csfd(-1), socket_fd(-1), notify_fd(-1),
-              force_notify_fd(-1), notify_var(nullptr), uid(uid), gid(gid), rlimits(rlimits)
+              force_notify_fd(-1), output_fd(-1), notify_var(nullptr), uid(uid), gid(gid), rlimits(rlimits)
     { }
 };
 
@@ -65,7 +68,7 @@ class base_process_service;
 class process_restart_timer : public eventloop_t::timer_impl<process_restart_timer>
 {
     public:
-    base_process_service * service;
+    base_process_service *service;
 
     explicit process_restart_timer(base_process_service *service_p)
         : service(service_p)
@@ -79,7 +82,7 @@ class process_restart_timer : public eventloop_t::timer_impl<process_restart_tim
 class exec_status_pipe_watcher : public eventloop_t::fd_watcher_impl<exec_status_pipe_watcher>
 {
     public:
-    base_process_service * service;
+    base_process_service *service;
     dasynq::rearm fd_event(eventloop_t &eloop, int fd, int flags) noexcept;
 
     exec_status_pipe_watcher(base_process_service * sr) noexcept : service(sr) { }
@@ -92,7 +95,7 @@ class exec_status_pipe_watcher : public eventloop_t::fd_watcher_impl<exec_status
 class stop_status_pipe_watcher : public eventloop_t::fd_watcher_impl<stop_status_pipe_watcher>
 {
     public:
-    process_service * service;
+    process_service *service;
     dasynq::rearm fd_event(eventloop_t &eloop, int fd, int flags) noexcept;
 
     stop_status_pipe_watcher(process_service * sr) noexcept : service(sr) { }
@@ -105,7 +108,7 @@ class stop_status_pipe_watcher : public eventloop_t::fd_watcher_impl<stop_status
 class ready_notify_watcher : public eventloop_t::fd_watcher_impl<ready_notify_watcher>
 {
     public:
-    base_process_service * service;
+    base_process_service *service;
     dasynq::rearm fd_event(eventloop_t &eloop, int fd, int flags) noexcept;
 
     ready_notify_watcher(base_process_service * sr) noexcept : service(sr) { }
@@ -118,7 +121,7 @@ class ready_notify_watcher : public eventloop_t::fd_watcher_impl<ready_notify_wa
 class service_child_watcher : public eventloop_t::child_proc_watcher_impl<service_child_watcher>
 {
     public:
-    base_process_service * service;
+    base_process_service *service;
     dasynq::rearm status_change(eventloop_t &eloop, pid_t child, int status) noexcept;
 
     service_child_watcher(base_process_service * sr) noexcept : service(sr) { }
@@ -131,7 +134,7 @@ class service_child_watcher : public eventloop_t::child_proc_watcher_impl<servic
 class stop_child_watcher : public eventloop_t::child_proc_watcher_impl<stop_child_watcher>
 {
     public:
-    process_service * service;
+    process_service *service;
     dasynq::rearm status_change(eventloop_t &eloop, pid_t child, int status) noexcept;
 
     stop_child_watcher(process_service * sr) noexcept : service(sr) { }
@@ -140,6 +143,19 @@ class stop_child_watcher : public eventloop_t::child_proc_watcher_impl<stop_chil
     void operator=(const service_child_watcher &) = delete;
 };
 
+class log_output_watcher : public eventloop_t::fd_watcher_impl<log_output_watcher>
+{
+    public:
+    base_process_service *service;
+
+    dasynq::rearm fd_event(eventloop_t &eloop, int fd, int flags) noexcept;
+
+    log_output_watcher(base_process_service * sr) noexcept : service(sr) { }
+
+    log_output_watcher(const ready_notify_watcher &) = delete;
+    void operator=(const ready_notify_watcher &) = delete;
+};
+
 // Base class for process-based services.
 class base_process_service : public service_record
 {
@@ -147,10 +163,7 @@ class base_process_service : public service_record
     friend class exec_status_pipe_watcher;
     friend class base_process_service_test;
     friend class ready_notify_watcher;
-
-    private:
-    // Re-launch process
-    void do_restart() noexcept;
+    friend class log_output_watcher;
 
     protected:
     ha_string program_name;          // storage for program/script and arguments
@@ -164,6 +177,15 @@ class base_process_service : public service_record
     string working_dir;       // working directory (or empty)
     string env_file;          // file with environment settings for this service
 
+    log_type_id log_type = log_type_id::NONE;
+    string logfile;          // log file name, empty string specifies /dev/null
+    int logfile_perms = 0;   // logfile permissions("mode")
+    uid_t logfile_uid = -1;  // logfile owner user id
+    gid_t logfile_gid = -1;  // logfile group id
+    unsigned log_buf_max = 0; // log buffer maximum size
+    unsigned log_buf_size = 0; // log buffer current size
+    std::vector<char, default_init_allocator<char>> log_buffer;
+
     std::vector<service_rlimits> rlimits; // resource limits
 
 #if SUPPORT_CGROUPS
@@ -173,6 +195,7 @@ class base_process_service : public service_record
     service_child_watcher child_listener;
     exec_status_pipe_watcher child_status_listener;
     process_restart_timer process_timer; // timer is used for start, stop and restart
+    log_output_watcher log_output_listener;
     time_val last_start_time;
 
     // Restart interval time and restart count are used to track the number of automatic restarts
@@ -202,6 +225,8 @@ class base_process_service : public service_record
     bp_sys::exit_status exit_status; // Exit status, if the process has exited (pid == -1).
     int socket_fd = -1;  // For socket-activation services, this is the file descriptor for the socket.
     int notification_fd = -1;  // If readiness notification is via fd
+    int log_output_fd = -1; // If logging via buffer/pipe, write end of the pipe
+    int log_input_fd = -1; // If logging via buffer/pipe, read end of the pipe
 
     // Only one of waiting_restart_timer and waiting_stopstart_timer should be set at any time.
     // They indicate that the process timer is armed (and why).
@@ -214,6 +239,11 @@ class base_process_service : public service_record
     // If executing child process failed, information about the error
     run_proc_err exec_err_info;
 
+    private:
+    // Re-launch process
+    void do_restart() noexcept;
+
+    protected:
     // Run a child process (call after forking). Note that some parameters specify file descriptors,
     // but in general file descriptors may be moved before the exec call.
     void run_child_proc(run_proc_params params) noexcept;
@@ -256,6 +286,10 @@ class base_process_service : public service_record
 
     void becoming_inactive() noexcept override;
 
+    // Get the file descriptor which the process should read input from (STDIN).
+    // Return false on failure. If input_fd returned is -1, process has no specific input.
+    virtual bool get_input_fd(int *input_fd) noexcept { *input_fd = -1; return true; }
+
     // Kill with SIGKILL
     virtual void kill_with_fire() noexcept;
 
@@ -271,6 +305,8 @@ class base_process_service : public service_record
         return nullptr;
     }
 
+    bool ensure_log_buffer_backing(unsigned size) noexcept;
+
     public:
     // Constructor for a base_process_service. Note that the various parameters not specified here must in
     // general be set separately (using the appropriate set_xxx function for each).
@@ -284,6 +320,7 @@ class base_process_service : public service_record
             child_listener.unreserve(event_loop);
         }
         process_timer.deregister(event_loop);
+        set_log_mode(log_type_id::NONE);
     }
 
     // Set the command to run this service (executable and arguments, nul separated). The command_parts_p
@@ -318,6 +355,94 @@ class base_process_service : public service_record
         stop_arg_parts = std::move(command_parts);
     }
 
+    void set_logfile_details(string &&logfile, int logfile_perms, uid_t logfile_uid, gid_t logfile_gid)
+            noexcept
+    {
+        this->logfile = std::move(logfile);
+        this->logfile_perms = logfile_perms;
+        this->logfile_uid = logfile_uid;
+        this->logfile_gid = logfile_gid;
+    }
+
+    // Set log buffer maximum size (for if mode is BUFFER). Maximum allowed size is UINT_MAX / 2
+    // (must be checked by caller).
+    void set_log_buf_max(unsigned max_size) noexcept
+    {
+        this->log_buf_max = max_size;
+    }
+
+    // Set log mode (NONE, BUFFER, FILE, PIPE) (must not change mode when service is not STOPPED).
+    void set_log_mode(log_type_id log_type) noexcept
+    {
+        if (this->log_type == log_type) {
+            return;
+        }
+        if (log_output_fd != -1) {
+            if (this->log_type == log_type_id::BUFFER) {
+                log_output_listener.deregister(event_loop);
+            }
+            if (!value(log_type).is_in(log_type_id::BUFFER, log_type_id::PIPE)) {
+                bp_sys::close(log_output_fd);
+                bp_sys::close(log_input_fd);
+                log_output_fd = log_input_fd = -1;
+            }
+        }
+        this->log_type = log_type;
+    }
+
+    log_type_id get_log_mode() noexcept
+    {
+        return this->log_type;
+    }
+
+    // Set the output pipe descriptors (both read and write end). This is only valid to call for
+    // log mode PIPE and only when the service has just been loaded. It is intended for transfer
+    // of fds from a placeholder service.
+    void set_output_pipe_fds(std::pair<int,int> fds)
+    {
+        log_input_fd = fds.first;
+        log_output_fd = fds.second;
+    }
+
+    std::pair<int,int> transfer_output_pipe() noexcept override
+    {
+        std::pair<int,int> r { log_input_fd, log_output_fd };
+        if (log_type == log_type_id::BUFFER && log_output_fd != -1) {
+            log_output_listener.deregister(event_loop);
+        }
+        log_input_fd = log_output_fd = -1;
+        return r;
+    }
+
+    int get_output_pipe_fd() noexcept override
+    {
+        if (log_input_fd != -1) {
+            return log_input_fd;
+        }
+
+        int pipefds[2];
+        if (bp_sys::pipe2(pipefds, O_CLOEXEC) == -1) {
+            log(loglevel_t::ERROR, get_name(), ": Can't open output pipe: ", std::strerror(errno));
+            return -1;
+        }
+        log_input_fd = pipefds[0];
+        log_output_fd = pipefds[1];
+        return log_input_fd;
+    }
+
+    // Get the log buffer (address, length)
+    std::pair<const char *, unsigned> get_log_buffer() noexcept
+    {
+        return {log_buffer.data(), log_buf_size};
+    }
+
+    void clear_log_buffer() noexcept
+    {
+        log_buffer.clear();
+        log_buffer.shrink_to_fit();
+        log_buf_size = 0;
+    }
+
     void set_env_file(const std::string &env_file_p)
     {
         env_file = env_file_p;
@@ -375,12 +500,8 @@ class base_process_service : public service_record
     }
 
     // Set the working directory
-    void set_working_dir(const string &working_dir_p)
-    {
-        working_dir = working_dir_p;
-    }
-
-    void set_working_dir(string &&working_dir_p) noexcept
+    // Note: constructing parameter may throw!
+    void set_working_dir(string working_dir_p) noexcept
     {
         working_dir = std::move(working_dir_p);
     }
@@ -450,9 +571,10 @@ class process_service : public base_process_service
     stop_child_watcher stop_watcher;
     stop_status_pipe_watcher stop_pipe_watcher;
 
-    protected:
     bool doing_smooth_recovery = false; // if we are performing smooth recovery
 
+    service_record *consumer_for = nullptr;
+
 #if USE_UTMPX
 
     private:
@@ -507,6 +629,7 @@ class process_service : public base_process_service
                 // Within the restart limiting interval; check number of restarts
                 if (restart_interval_count >= max_restart_interval_count) {
                     log(loglevel_t::ERROR, "Service ", get_name(), " restarting too quickly; stopping.");
+                    set_target_state(service_state_t::STOPPED);
                     return false;
                 }
                 ++restart_interval_count;
@@ -521,6 +644,19 @@ class process_service : public base_process_service
         return true;
     }
 
+    bool get_input_fd(int *input_fd) noexcept override
+    {
+        if (consumer_for == nullptr) {
+            *input_fd = -1;
+            return true;
+        }
+
+        int cfd = consumer_for->get_output_pipe_fd();
+        if (cfd == -1) return false;
+        *input_fd = cfd;
+        return true;
+    }
+
     process_service(service_set *sset, const string &name, service_type_t s_type, ha_string &&command,
             std::list<std::pair<unsigned,unsigned>> &command_offsets,
             const std::list<prelim_dep> &depends_p)
@@ -571,11 +707,25 @@ class process_service : public base_process_service
 
 #endif
 
+    // Set this service to consume output of specified service (only call while service stopped).
+    void set_consumer_for(service_record *consumed) noexcept
+    {
+        consumer_for = consumed;
+    }
+
+    service_record *get_consumed() noexcept
+    {
+        return consumer_for;
+    }
+
     ~process_service() noexcept
     {
         if (reserved_stop_watch) {
             stop_watcher.unreserve(event_loop);
         }
+        if (consumer_for != nullptr) {
+            consumer_for->set_log_consumer(nullptr);
+        }
     }
 };
 
