@@ -8,17 +8,17 @@
 #include <unordered_set>
 #include <algorithm>
 
-#include "dasynq.h"
+#include <dasynq.h>
 
-#include "dinit.h"
-#include "control.h"
-#include "service-listener.h"
-#include "service-constants.h"
-#include "load-service.h"
-#include "dinit-ll.h"
-#include "dinit-log.h"
-#include "service-dir.h"
-#include "dinit-env.h"
+#include <dinit.h>
+#include <control.h>
+#include <service-listener.h>
+#include <service-constants.h>
+#include <load-service.h>
+#include <dinit-ll.h>
+#include <dinit-log.h>
+#include <service-dir.h>
+#include <dinit-env.h>
 
 /*
  * This header defines service_record, a data record maintaining information about a service,
@@ -147,6 +147,7 @@
 class service_record;
 class service_set;
 class base_process_service;
+class process_service;
 
 /* Service dependency record */
 class service_dep
@@ -196,6 +197,11 @@ class service_dep
     {
         to = new_to;
     }
+
+    void set_from(service_record *new_from) noexcept
+    {
+        from = new_from;
+    }
 };
 
 /* preliminary service dependency information */
@@ -215,11 +221,19 @@ class prelim_dep
 inline void log_service_load_failure(service_description_exc &exc)
 {
     if (exc.line_num != (unsigned)-1) {
-        log(loglevel_t::ERROR, "Couldn't load service '", exc.service_name, "' (line ", exc.line_num, "): ",
-                exc.exc_description);
+        if (exc.setting_name == nullptr) {
+            log(loglevel_t::ERROR, "Error in service description for '", exc.service_name, "' (line ", exc.line_num, "): ",
+                    exc.exc_description);
+        }
+        else {
+            log(loglevel_t::ERROR, "Error in service description for '", exc.service_name, "': setting '", exc.setting_name, "' "
+                    "(on line ", exc.line_num, "): ",
+                    exc.exc_description);
+        }
     }
     else {
-        log(loglevel_t::ERROR, "Couldn't load service '", exc.service_name, "' setting '", exc.setting_name, "': ",
+        // If no line number, setting name must be present
+        log(loglevel_t::ERROR, "Error in service description for '", exc.service_name, "' setting '", exc.setting_name, "': ",
                 exc.exc_description);
     }
 }
@@ -239,7 +253,7 @@ class service_record
     
     private:
     string service_name;
-    service_type_t record_type;  // service_type_t::DUMMY, PROCESS, SCRIPTED, or INTERNAL
+    service_type_t record_type;
 
     // 'service_state' can be any valid state: STARTED, STARTING, STOPPING, STOPPED.
     // 'desired_state' is only set to final states: STARTED or STOPPED.
@@ -248,9 +262,10 @@ class service_record
 
     protected:
     service_flags_t onstart_flags;
-
-    string logfile;           // log file name, empty string specifies /dev/null
     
+    environment service_env; // holds the environment populated during load
+    const char *service_dsc_dir = nullptr; // directory containing service description file
+
     bool auto_restart : 1;    // whether to restart this (process) if it dies unexpectedly
     bool smooth_recovery : 1; // whether the service process can restart without bringing down service
 
@@ -280,6 +295,8 @@ class service_record
     bool in_auto_restart : 1;
     bool in_user_restart : 1;
 
+    bool is_loading : 1;        // used to detect cyclic dependencies when loading a service
+
     int required_by = 0;        // number of dependents wanting this service to be started
 
     // list of dependencies
@@ -295,6 +312,8 @@ class service_record
     
     std::unordered_set<service_listener *> listeners;
     
+    process_service *log_consumer = nullptr;
+
     // Process services:
     bool force_stop; // true if the service must actually stop. This is the
                      // case if for example the process dies; the service,
@@ -303,7 +322,7 @@ class service_record
     int term_signal = SIGTERM;  // signal to use for process termination
     
     string socket_path; // path to the socket for socket-activation service
-    int socket_perms;   // socket permissions ("mode")
+    int socket_perms = 0;   // socket permissions ("mode")
     uid_t socket_uid = -1;  // socket user id or -1
     gid_t socket_gid = -1;  // socket group id or -1
 
@@ -393,9 +412,6 @@ class service_record
     // Release console (console must be currently held by this service)
     void release_console() noexcept;
     
-    // Started state reached
-    bool process_started() noexcept;
-
     // Initiate definite startup
     void initiate_start() noexcept;
 
@@ -411,11 +427,16 @@ class service_record
         service_state = new_state;
     }
 
+    // Set the target state
+    void set_target_state(service_state_t new_target_state) noexcept
+    {
+        desired_state = new_target_state;
+    }
+
     // Virtual functions, to be implemented by service implementations:
 
     // Do any post-dependency startup; return false on failure. Should return true if service
-    // has started or is in the process of starting. Will only be called once can_proceed_to_start()
-    // returns true.
+    // has started or is in the process of starting.
     virtual bool bring_up() noexcept;
 
     // All dependents have stopped, and this service should proceed to stop.
@@ -445,6 +466,9 @@ class service_record
 
     public:
 
+    class loading_tag_cls { };
+    static const loading_tag_cls LOADING_TAG;
+
     service_record(service_set *set, const string &name)
         : service_name(name), service_state(service_state_t::STOPPED),
             desired_state(service_state_t::STOPPED), auto_restart(false), smooth_recovery(false),
@@ -453,11 +477,16 @@ class service_record
             waiting_for_execstat(false), start_explicit(false), prop_require(false), prop_release(false),
             prop_failure(false), prop_start(false), prop_stop(false), prop_pin_dpt(false),
             start_failed(false), start_skipped(false), in_auto_restart(false), in_user_restart(false),
-            force_stop(false)
+            is_loading(false), force_stop(false)
     {
         services = set;
-        record_type = service_type_t::DUMMY;
-        socket_perms = 0;
+        record_type = service_type_t::PLACEHOLDER;
+    }
+
+    // Construct a service record and mark it as a placeholder for a loading service
+    service_record(service_set *set, const string &name, loading_tag_cls tag)
+        : service_record(set, name) {
+        is_loading = true;
     }
 
     service_record(service_set *set, const string &name, service_type_t record_type_p,
@@ -522,15 +551,20 @@ class service_record
         return start_explicit;
     }
 
-    // Set logfile, should be done before service is started
-    void set_log_file(const string &logfile)
+    // Set directory containing service description file
+    void set_service_dsc_dir(const char *dsc_dir) noexcept
     {
-        this->logfile = logfile;
+        service_dsc_dir = dsc_dir;
     }
-    
-    void set_log_file(std::string &&logfile) noexcept
+
+    const char *get_service_dsc_dir() noexcept
     {
-        this->logfile = std::move(logfile);
+        return service_dsc_dir;
+    }
+
+    void set_environment(environment &&env) noexcept
+    {
+        this->service_env = std::move(env);
     }
 
     // Set whether this service should automatically restart when it dies
@@ -570,6 +604,16 @@ class service_record
         start_on_completion = std::move(chain_to);
     }
 
+    void set_log_consumer(process_service *consumer)
+    {
+        log_consumer = consumer;
+    }
+
+    process_service *get_log_consumer()
+    {
+        return log_consumer;
+    }
+
     const std::string &get_name() const noexcept { return service_name; }
     service_state_t get_state() const noexcept { return service_state; }
     
@@ -604,9 +648,9 @@ class service_record
     }
 
     // Is this a dummy service (used only when loading a new service)?
-    bool is_dummy() noexcept
+    bool check_is_loading() noexcept
     {
-        return record_type == service_type_t::DUMMY;
+        return is_loading;
     }
     
     bool did_start_fail() noexcept
@@ -632,14 +676,17 @@ class service_record
     }
     
     // Assuming there is one reference (from a control link), return true if this is the only reference,
-    // or false if there are others (including dependents).
+    // or false if there are others (including dependents, excluding dependents via "before" and "after"
+    // links).
     bool has_lone_ref(bool check_deps = true) noexcept
     {
         if (check_deps) {
             for (auto *dept : dependents) {
                 // BEFORE links don't count because they are actually specified via the "to" service i.e.
-                // this service.
-                if (dept->dep_type != dependency_type::BEFORE) {
+                // this service. AFTER links don't count because, although they come from the dependent,
+                // they do not reflect an actual dependency, just an ordering, and this service can still
+                // be unloaded if they exist (and replaced with a placeholder).
+                if (!value(dept->dep_type).is_in(dependency_type::BEFORE, dependency_type::AFTER)) {
                     return false;
                 }
             }
@@ -648,23 +695,24 @@ class service_record
         return (++i == listeners.end());
     }
 
-    // Prepare this service to be unloaded.
-    void prepare_for_unload() noexcept
+    // Check whether this service has no dependents/dependencies/references that would keep it from
+    // unloading. Does not check listeners (i.e. mostly useful for placeholder services).
+    bool is_unrefd() noexcept
     {
-        // Remove all dependencies:
-        for (auto &dep : depends_on) {
-            auto &dep_dpts = dep.get_to()->dependents;
-            dep_dpts.erase(std::find(dep_dpts.begin(), dep_dpts.end(), &dep));
-        }
-        depends_on.clear();
-
-        // Also remove all dependents. This should not be necessary except for "before" links.
-        // Note: this for loop might look odd, but it's correct!
-        for (auto i = dependents.begin(); i != dependents.end(); i = dependents.begin()) {
-            (*i)->get_from()->rm_dep(**i);
-        }
+        return depends_on.empty() && dependents.empty() && log_consumer == nullptr;
     }
 
+    // Get fd corresponding to the read end of the pipe/socket connected to the write end used by the
+    // service process (if any). Opens the pipe if not already open; returns -1 if the service output
+    // cannot be piped (failure to open pipe, wrong service type, etc).
+    virtual int get_output_pipe_fd() noexcept
+    {
+        return -1;
+    }
+
+    // Prepare this service to be unloaded.
+    void prepare_for_unload() noexcept;
+
     // Why did the service stop?
     stopped_reason_t get_stop_reason()
     {
@@ -781,13 +829,95 @@ class service_record
     // Start a specific dependency of this service. Should only be called if this service is in an
     // appropriate state (started, starting). The dependency is marked as holding acquired; when
     // this service stops, the dependency will be released and may also stop.
-    void start_dep(service_dep &dep)
+    void start_dep(service_dep &dep) noexcept
     {
-        if (! dep.holding_acq) {
+        if (!dep.holding_acq) {
             dep.get_to()->require();
             dep.holding_acq = true;
         }
     }
+
+    // Transfer the file descriptors representing the output (logging) pipe for this service. The file
+    // descriptors are returned as a pair (read,write) (possibly with -1,-1 if there is no log pipe open)
+    // and disassociated from this service.
+    // This is used when a process is reloaded (for example) to transfer the pipe from the original service
+    // record to the new service record.
+    virtual std::pair<int,int> transfer_output_pipe() noexcept
+    {
+        return {-1,-1};
+    }
+};
+
+class placeholder_service : public service_record {
+    int log_output_fd = -1; // write end of the output pipe
+    int log_input_fd = -1; // read end of the output pipe
+
+    public:
+    placeholder_service(service_set *set, const string &name)
+        : service_record(set, name, service_type_t::PLACEHOLDER, {}) {
+    }
+
+    int get_output_pipe_fd() noexcept override
+    {
+        if (log_input_fd != -1) {
+            return log_input_fd;
+        }
+
+        int pipefds[2];
+        if (bp_sys::pipe2(pipefds, O_CLOEXEC) == -1) {
+            log(loglevel_t::ERROR, get_name(), " (placeholder): Can't open output pipe: ", std::strerror(errno));
+            return -1;
+        }
+        log_input_fd = pipefds[0];
+        log_output_fd = pipefds[1];
+        return log_input_fd;
+    }
+
+    std::pair<int,int> transfer_output_pipe() noexcept override
+    {
+        std::pair<int,int> r { log_input_fd, log_output_fd };
+        log_input_fd = log_output_fd = -1;
+        return r;
+    }
+
+    void set_output_pipe(std::pair<int,int> fds) noexcept
+    {
+        log_input_fd = fds.first;
+        log_output_fd = fds.second;
+    }
+
+    ~placeholder_service()
+    {
+        if (log_output_fd != -1) {
+            bp_sys::close(log_output_fd);
+            bp_sys::close(log_input_fd);
+        }
+    }
+};
+
+// Externally triggered service. This is essentially the same as an internal service, but does not start
+// until the external trigger is set.
+class triggered_service : public service_record {
+    private:
+    bool is_triggered = false;
+
+    public:
+    using service_record::service_record;
+
+    bool bring_up() noexcept override;
+
+    bool can_interrupt_start() noexcept override
+    {
+        return true;
+    }
+
+    void set_trigger(bool new_trigger) noexcept
+    {
+        is_triggered = new_trigger;
+        if (is_triggered && get_state() == service_state_t::STARTING && !waiting_for_deps) {
+            started();
+        }
+    }
 };
 
 inline auto extract_prop_queue(service_record *sr) -> decltype(sr->prop_queue_node) &
@@ -868,7 +998,7 @@ class service_set
     }
 
     // Locate an existing service record.
-    service_record *find_service(const std::string &name) noexcept;
+    service_record *find_service(const std::string &name, bool find_placeholders = false) noexcept;
 
     // Load a service description, and dependencies, if there is no existing
     // record for the given name.
@@ -885,7 +1015,7 @@ class service_set
     }
 
     // Re-load a service description from file. If the service type changes then this returns
-    // a new service instead (the old one should be removed and deleted by the caller).
+    // a new service instead (the old pointer is no longer valid).
     // Throws:
     //   service_load_exc (or subclass) on problem with service description
     //   std::bad_alloc on out-of-memory condition
@@ -923,6 +1053,60 @@ class service_set
         *i = replacement;
     }
 
+    // Unload a service, possibly replacing it with a placeholder service. This can fail with bad_alloc.
+    // svc is no longer valid on return.
+    void unload_service(service_record *svc)
+    {
+        // Check if there are any "after" dependents, or any "before" dependencies. We need a placeholder
+        // if so. Note that the placeholder is created before making any destructive changes, so if we fail
+        // to create it no rollback is needed.
+        placeholder_service *placeholder = nullptr;
+        auto make_placeholder = [&]() {
+            if (placeholder == nullptr) {
+                std::unique_ptr<placeholder_service> ph { new placeholder_service(this, svc->get_name()) };
+                add_service(ph.get());
+                placeholder = ph.release();
+            }
+        };
+
+        auto *consumer = svc->get_log_consumer();
+        if (consumer != nullptr) {
+            make_placeholder();
+            placeholder->set_output_pipe(svc->transfer_output_pipe());
+        }
+
+        auto &svc_depts = svc->get_dependents();
+        for (auto i = svc_depts.begin(); i != svc_depts.end(); ) {
+            auto *dept = *i;
+            if (dept->dep_type == dependency_type::AFTER) {
+                make_placeholder();
+                dept->set_to(placeholder);
+                auto &ph_depts = placeholder->get_dependents();
+                ph_depts.splice(ph_depts.end(), svc_depts, i++);
+                continue;
+            }
+            ++i;
+        }
+
+        auto &svc_deps = svc->get_dependencies();
+        for (auto i = svc_deps.begin(); i != svc_deps.end(); ) {
+            auto &dep = *i;
+            if (dep.dep_type == dependency_type::BEFORE) {
+                make_placeholder();
+                dep.set_from(placeholder);
+                auto &ph_deps = placeholder->get_dependencies();
+                ph_deps.splice(ph_deps.end(), svc_deps, i++);
+                continue;
+            }
+            ++i;
+        }
+
+        // Proceed with unload.
+        svc->prepare_for_unload();
+        remove_service(svc);
+        delete svc;
+    }
+
     // Get the list of all loaded services.
     const std::list<service_record *> &list_services() noexcept
     {
