@@ -67,8 +67,8 @@ void process_service::exec_succeeded() noexcept
         }
     }
     else if (get_state() == service_state_t::STARTED) {
-        // Smooth recovery (is now complete)
-        if (waiting_stopstart_timer) {
+        // Smooth recovery (is now complete, if we don't need readiness notification)
+        if (waiting_stopstart_timer && notification_fd == -1) {
             process_timer.stop_timer(event_loop);
             waiting_stopstart_timer = false;
         }
@@ -85,8 +85,8 @@ void process_service::exec_succeeded() noexcept
 
 void scripted_service::exec_succeeded() noexcept
 {
-	// For a scripted service, this means nothing other than that the start/stop
-	// script will now begin.
+    // For a scripted service, this means nothing other than that the start/stop
+    // script will now begin.
 }
 
 rearm exec_status_pipe_watcher::fd_event(eventloop_t &loop, int fd, int flags) noexcept
@@ -257,11 +257,68 @@ dasynq::rearm stop_child_watcher::status_change(eventloop_t &loop, pid_t child,
     return dasynq::rearm::NOOP;
 }
 
+rearm log_output_watcher::fd_event(eventloop_t &eloop, int fd, int flags) noexcept
+{
+    // In case buffer size has been decreased, check if we are already at the limit:
+    if (service->log_buf_size >= service->log_buf_max) {
+        // If so, read and discard.
+        char buf[1024];
+        int r = bp_sys::read(fd, buf, 1024);
+        if (r == -1) {
+            if (errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR) {
+                return rearm::REARM;
+            }
+            goto bad_read;
+        }
+        if (r == 0) goto eof_read;
+        return rearm::REARM;
+    }
+
+    {
+        size_t max_read = std::max(service->log_buf_max / 8, 256u);
+        max_read = std::min((unsigned)max_read, service->log_buf_max - service->log_buf_size);
+
+        // ensure vector has size sufficient to read
+        unsigned new_size = service->log_buf_size + max_read;
+        if (!service->ensure_log_buffer_backing(new_size)) {
+            return rearm::DISARM;
+        }
+
+        max_read = service->log_buffer.size() - service->log_buf_size;
+
+        int r = bp_sys::read(fd, service->log_buffer.data() + service->log_buf_size, max_read);
+        if (r == -1) {
+            if (errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR) {
+                return rearm::REARM;
+            }
+            goto bad_read;
+        }
+
+        if (r == 0) goto eof_read;
+
+        service->log_buf_size += r;
+        return rearm::REARM;
+    }
+
+    // error/end-of-stream handling:
+
+    bad_read:
+    log(loglevel_t::WARN, "Service ", service->get_name(), " output not readable: ", strerror(errno));
+
+    eof_read:
+    deregister(eloop);
+    close(fd);
+    close(service->log_output_fd);
+    service->log_input_fd = -1;
+    service->log_output_fd = -1;
+    return rearm::REMOVED;
+}
+
 void process_service::handle_exit_status(bp_sys::exit_status exit_status) noexcept
 {
     bool did_exit = exit_status.did_exit();
     bool was_signalled = exit_status.was_signalled();
-    auto service_state = get_state();
+    auto current_state = get_state();
 
     if (notification_fd != -1) {
         readiness_watcher.deregister(event_loop);
@@ -269,7 +326,7 @@ void process_service::handle_exit_status(bp_sys::exit_status exit_status) noexce
         notification_fd = -1;
     }
 
-    if (!exit_status.did_exit_clean() && service_state != service_state_t::STOPPING) {
+    if (!exit_status.did_exit_clean() && current_state != service_state_t::STOPPING) {
         if (did_exit) {
             log(loglevel_t::ERROR, "Service ", get_name(), " process terminated with exit code ",
                     exit_status.get_exit_status());
@@ -283,6 +340,11 @@ void process_service::handle_exit_status(bp_sys::exit_status exit_status) noexce
     if (waiting_stopstart_timer) {
         process_timer.stop_timer(event_loop);
         waiting_stopstart_timer = false;
+        if (current_state == service_state_t::STARTED) {
+            // Must have been in smooth recovery and waiting for readiness notification.
+            // Treat this the same as if we were STARTING:
+            current_state = service_state_t::STARTING;
+        }
     }
 
 #if USE_UTMPX
@@ -291,14 +353,13 @@ void process_service::handle_exit_status(bp_sys::exit_status exit_status) noexce
     }
 #endif
 
-    if (service_state == service_state_t::STARTING) {
+    if (current_state == service_state_t::STARTING) {
         // If state is STARTING, we must be waiting for readiness notification; the process has
         // terminated before becoming ready.
         stop_reason = stopped_reason_t::FAILED;
-        service_state = service_state_t::STOPPING;
         failed_to_start();
     }
-    else if (service_state == service_state_t::STOPPING) {
+    else if (current_state == service_state_t::STOPPING) {
         // We won't log a non-zero exit status or termination due to signal here -
         // we assume that the process died because we signalled it.
         if (waiting_stopstart_timer) {
@@ -314,7 +375,7 @@ void process_service::handle_exit_status(bp_sys::exit_status exit_status) noexce
             initiate_start();
         }
     }
-    else if (smooth_recovery && service_state == service_state_t::STARTED && check_restart()) {
+    else if (smooth_recovery && current_state == service_state_t::STARTED && check_restart()) {
         // unexpected termination, with smooth recovery
         doing_smooth_recovery = true;
         do_smooth_recovery();
@@ -367,9 +428,9 @@ void bgproc_service::handle_exit_status(bp_sys::exit_status exit_status) noexcep
     begin:
     bool did_exit = exit_status.did_exit();
     bool was_signalled = exit_status.was_signalled();
-    auto service_state = get_state();
+    auto current_state = get_state();
 
-    if (!exit_status.did_exit_clean() && service_state != service_state_t::STOPPING) {
+    if (!exit_status.did_exit_clean() && current_state != service_state_t::STOPPING) {
         if (did_exit) {
             log(loglevel_t::ERROR, "Service ", get_name(), " process terminated with exit code ",
                     exit_status.get_exit_status());
@@ -390,7 +451,7 @@ void bgproc_service::handle_exit_status(bp_sys::exit_status exit_status) noexcep
 
         // We're either started, or stopping (i.e. we were requested to stop during smooth recovery).
 
-        if (service_state == service_state_t::STOPPING) {
+        if (current_state == service_state_t::STOPPING) {
             // Stop was issued during smooth recovery
             if ((did_exit && exit_status.get_exit_status() != 0) || was_signalled) {
                 if (!waiting_for_deps) {
@@ -460,14 +521,13 @@ void bgproc_service::handle_exit_status(bp_sys::exit_status exit_status) noexcep
         }
     }
 
-    if (service_state == service_state_t::STARTING) {
+    if (current_state == service_state_t::STARTING) {
         if (exit_status.did_exit_clean()) {
             auto pid_result = read_pid_file(&exit_status);
             switch (pid_result) {
                 case pid_result_t::FAILED:
                     // Failed startup: no auto-restart.
                     stop_reason = stopped_reason_t::FAILED;
-                    service_state = service_state_t::STOPPING;
                     failed_to_start();
                     break;
                 case pid_result_t::TERMINATED:
@@ -481,11 +541,10 @@ void bgproc_service::handle_exit_status(bp_sys::exit_status exit_status) noexcep
         }
         else {
             stop_reason = stopped_reason_t::FAILED;
-            service_state = service_state_t::STOPPING;
             failed_to_start();
         }
     }
-    else if (service_state == service_state_t::STOPPING) {
+    else if (current_state == service_state_t::STOPPING) {
         // We won't log a non-zero exit status or termination due to signal here -
         // we assume that the process died because we signalled it.
         if (stop_pid == -1 && !waiting_for_execstat) {
@@ -534,14 +593,14 @@ void scripted_service::handle_exit_status(bp_sys::exit_status exit_status) noexc
 {
     bool did_exit = exit_status.did_exit();
     bool was_signalled = exit_status.was_signalled();
-    auto service_state = get_state();
+    auto current_state = get_state();
 
     // For a scripted service, a termination occurs in one of three main cases:
     // - the start script completed (or failed), when service was STARTING
     // - the start script was interrupted to cancel startup; state is STOPPING
     // - the stop script complete (or failed), state is STOPPING
 
-    if (service_state == service_state_t::STOPPING) {
+    if (current_state == service_state_t::STOPPING) {
         // We might be running the stop script, or we might be running the start script and have issued
         // a cancel order via SIGINT:
         if (interrupting_start) {
@@ -607,7 +666,6 @@ void scripted_service::handle_exit_status(bp_sys::exit_status exit_status) noexc
                 log(loglevel_t::ERROR, "Service ", get_name(), " command terminated due to signal ",
                         exit_status.get_term_sig());
             }
-            service_state = service_state_t::STOPPED;
             stop_reason = stopped_reason_t::FAILED;
             failed_to_start();
         }
@@ -619,13 +677,12 @@ void scripted_service::exec_failed(run_proc_err errcode) noexcept
 {
     log(loglevel_t::ERROR, get_name(), ": execution failed - ",
             exec_stage_descriptions[static_cast<int>(errcode.stage)], ": ", strerror(errcode.st_errno));
-    auto service_state = get_state();
-    if (service_state == service_state_t::STARTING) {
+    auto current_state = get_state();
+    if (current_state == service_state_t::STARTING) {
         stop_reason = stopped_reason_t::EXECFAILED;
-        service_state = service_state_t::STOPPING;
         failed_to_start();
     }
-    else if (service_state == service_state_t::STOPPING) {
+    else if (current_state == service_state_t::STOPPING) {
         // We've logged the failure, but it's probably better not to leave the service in
         // STOPPING state:
         stopped();
@@ -789,10 +846,10 @@ void process_service::kill_with_fire() noexcept
 
 void scripted_service::bring_down() noexcept
 {
-	if (pid != -1) {
-		// We're already running the stop script; nothing to do.
-		return;
-	}
+    if (pid != -1) {
+        // We're already running the stop script; nothing to do.
+        return;
+    }
 
     if (stop_command.length() == 0) {
         stopped();
@@ -861,7 +918,7 @@ bool process_service::start_stop_process(const std::vector<const char *> &cmd) n
 
     if (forkpid == 0) {
         close(pipefd[0]);
-        const char * working_dir_c = nullptr;
+        const char * working_dir_c = service_dsc_dir;
         if (! working_dir.empty()) working_dir_c = working_dir.c_str();
         run_proc_params run_params{cmd.data(), working_dir_c, logfile, pipefd[1], run_as_uid, run_as_gid, rlimits};
         run_params.on_console = false;
