@@ -20,7 +20,7 @@
  * See service.h for details.
  */
 
-// Find the requested service by name
+// Find the requested service by name.
 static service_record * find_service(const std::list<service_record *> & records,
                                     const char *name) noexcept
 {
@@ -34,9 +34,42 @@ static service_record * find_service(const std::list<service_record *> & records
     return nullptr;
 }
 
-service_record * service_set::find_service(const std::string &name) noexcept
+service_record * service_set::find_service(const std::string &name, bool find_placeholders) noexcept
 {
-    return ::find_service(records, name.c_str());
+    service_record *r = ::find_service(records, name.c_str());
+    if (r != nullptr && !find_placeholders && r->get_type() == service_type_t::PLACEHOLDER) {
+        return nullptr;
+    }
+    return r;
+}
+
+void service_record::prepare_for_unload() noexcept
+{
+    // Remove all dependencies:
+    for (auto &dep : depends_on) {
+        service_record *dependency = dep.get_to();
+        auto &dep_dpts = dependency->dependents;
+        dep_dpts.erase(std::find(dep_dpts.begin(), dep_dpts.end(), &dep));
+        if (dep.dep_type == dependency_type::AFTER) {
+            if (dependency->get_type() == service_type_t::PLACEHOLDER) {
+                if (dependency->is_unrefd()) {
+                    services->remove_service(dependency);
+                    delete dependency;
+                }
+            }
+        }
+    }
+    depends_on.clear();
+
+    // Also remove all dependents. This should not be necessary except for "before" links.
+    for (auto i = dependents.begin(); i != dependents.end();) {
+        service_record *before_svc = (*i)->get_from();
+        before_svc->rm_dep(**i++);
+        if (before_svc->get_type() == service_type_t::PLACEHOLDER && before_svc->is_unrefd()) {
+            services->remove_service(before_svc);
+            delete before_svc;
+        }
+    }
 }
 
 // Called when a service has actually stopped; dependents have stopped already, unless this stop
@@ -160,13 +193,13 @@ void service_record::release(bool issue_stop) noexcept
         // a require was pending though:
         prop_release = !prop_require;
         prop_require = false;
-        if (prop_release && service_state != service_state_t::STOPPED) {
+        if (prop_release) {
             services->add_prop_queue(this);
         }
 
         if (service_state != service_state_t::STOPPED && service_state != service_state_t::STOPPING
                 && issue_stop) {
-        	stop_reason = stopped_reason_t::NORMAL;
+            stop_reason = stopped_reason_t::NORMAL;
             do_stop();
         }
     }
@@ -208,7 +241,7 @@ void service_record::initiate_start() noexcept
     waiting_for_deps = true;
 
     if (start_check_dependencies()) {
-        waiting_for_deps = false;
+        // Note: we can't set waiting_for_deps false here since ordering dependencies might still kick in
         services->add_transition_queue(this);
     }
 }
@@ -280,12 +313,19 @@ void service_record::execute_transition() noexcept
 {
     if (service_state == service_state_t::STARTING) {
         if (check_deps_started()) {
+            waiting_for_deps = false;
             all_deps_started();
         }
     }
     else if (service_state == service_state_t::STOPPING) {
         if (stop_check_dependents()) {
             waiting_for_deps = false;
+            if (onstart_flags.kill_all_on_stop) {
+                log(loglevel_t::NOTICE, true, "Sending TERM/KILL to all processes...\n");
+                kill(-1, SIGTERM);
+                sleep(1);
+                kill(-1, SIGKILL);
+            }
             bring_down();
         }
     }
@@ -370,6 +410,15 @@ bool service_record::start_check_dependencies() noexcept
         }
     }
 
+    for (auto * dept : dependents) {
+        if (!dept->waiting_on && dept->is_only_ordering()) {
+            service_record *from = dept->get_from();
+            if (from->get_state() == service_state_t::STARTING) {
+                dept->waiting_on = true;
+            }
+        }
+    }
+
     return all_deps_started;
 }
 
@@ -386,7 +435,7 @@ bool service_record::check_deps_started() noexcept
 
 void service_record::all_deps_started() noexcept
 {
-    if (onstart_flags.starts_on_console && ! have_console) {
+    if (onstart_flags.starts_on_console && !have_console) {
         queue_for_console();
         return;
     }
@@ -420,7 +469,7 @@ void service_record::acquired_console() noexcept
 void service_record::started() noexcept
 {
     // If we start on console but don't keep it, release it now:
-    if (have_console && ! onstart_flags.runs_on_console) {
+    if (have_console && !onstart_flags.runs_on_console) {
         bp_sys::tcsetpgrp(0, bp_sys::getpgrp());
         release_console();
     }
@@ -497,7 +546,7 @@ void service_record::failed_to_start(bool depfailed, bool immediate_stop) noexce
     }
 
     start_failed = true;
-    log_service_failed(get_name());
+    log_service_failed(get_name(), depfailed);
     notify_listeners(service_event_t::FAILEDSTART);
     pinned_started = false;
 
@@ -569,7 +618,7 @@ void service_record::stop(bool bring_down) noexcept
     if (bring_down && service_state != service_state_t::STOPPED) {
         // Note even if we are already STOPPING we should call do_stop for the case where we need
         // to interrupt any currently ongoing restart of dependents.
-    	stop_reason = stopped_reason_t::NORMAL;
+        stop_reason = stopped_reason_t::NORMAL;
         do_stop();
     }
 }
@@ -626,14 +675,14 @@ void service_record::do_stop(bool with_restart) noexcept
         if (service_state == service_state_t::STARTING) {
             // If waiting for a dependency, or waiting for the console, we can interrupt start. Otherwise,
             // we need to delegate to can_interrupt_start() (which can be overridden).
-            if (! waiting_for_deps && ! waiting_for_console) {
-                if (! can_interrupt_start()) {
+            if (!waiting_for_deps && !waiting_for_console) {
+                if (!can_interrupt_start()) {
                     // Well this is awkward: we're going to have to continue starting. We can stop once
                     // we've reached the started state.
                     return;
                 }
 
-                if (! interrupt_start()) {
+                if (!interrupt_start()) {
                     // Now wait for service startup to actually end; we don't need to handle it here.
                     notify_listeners(service_event_t::STARTCANCELLED);
                     return;
@@ -699,7 +748,14 @@ bool service_record::stop_dependents(bool for_restart, bool restart_deps) noexce
 
             if (force_stop) {
                 // If this service is to be forcefully stopped, dependents must also be.
-                dep_from->forced_stop();
+                if (desired_state == service_state_t::STOPPED) {
+                    // If our target state was forced to STOPPED, this is a failure
+                    dep_from->stop_reason = stopped_reason_t::DEPFAILED;
+                    dep_from->unrecoverable_stop();
+                }
+                else {
+                    dep_from->forced_stop();
+                }
             }
 
             if (dep_from->get_state() != service_state_t::STOPPED) {
@@ -830,3 +886,11 @@ void service_set::service_inactive(service_record *sr) noexcept
 {
     active_services--;
 }
+
+bool triggered_service::bring_up() noexcept
+{
+    if (is_triggered) {
+        started();
+    }
+    return true;
+}
