@@ -4,19 +4,28 @@
 #include <string>
 #include <set>
 
-#include "dinit.h"
-#include "service.h"
-#include "baseproc-sys.h"
-#include "control.h"
+#include <dinit.h>
+#include <service.h>
+#include <baseproc-sys.h>
+#include <proc-service.h>
+#include <control.h>
+#include "control-datatypes.h"
 
 #include "../test_service.h"
+#include "../test_procservice.h"
 
 // Control protocol tests.
+#ifdef NDEBUG
+#error "This file must be built with assertions ENABLED!"
+#endif
+
+// common communication datatypes
+using namespace dinit_cptypes;
 
 class control_conn_t_test
 {
     public:
-    static service_record * service_from_handle(control_conn_t *cc, control_conn_t::handle_t handle)
+    static service_record * service_from_handle(control_conn_t *cc, handle_t handle)
     {
         return cc->find_service_for_key(handle);
     }
@@ -27,90 +36,90 @@ constexpr static int STATUS_BUFFER_SIZE = 6 + ((sizeof(pid_t) > sizeof(int)) ? s
 
 void cptest_queryver()
 {
-	service_set sset;
-	int fd = bp_sys::allocfd();
-	auto *cc = new control_conn_t(event_loop, &sset, fd);
+    service_set sset;
+    int fd = bp_sys::allocfd();
+    auto *cc = new control_conn_t(event_loop, &sset, fd);
 
-	bp_sys::supply_read_data(fd, { DINIT_CP_QUERYVERSION });
+    bp_sys::supply_read_data(fd, { (char)cp_cmd::QUERYVERSION });
 
-	event_loop.regd_bidi_watchers[fd]->read_ready(event_loop, fd);
+    event_loop.regd_bidi_watchers[fd]->read_ready(event_loop, fd);
 
-	// Write will process immediately, so there's no need for this:
-	//event_loop.regd_bidi_watchers[fd]->write_ready(event_loop, fd);
+    // Write will process immediately, so there's no need for this:
+    //event_loop.regd_bidi_watchers[fd]->write_ready(event_loop, fd);
 
-	// We expect a version number back:
-	std::vector<char> wdata;
-	bp_sys::extract_written_data(fd, wdata);
+    // We expect a version number back:
+    std::vector<char> wdata;
+    bp_sys::extract_written_data(fd, wdata);
 
-	assert(wdata.size() == 5);
-	assert(wdata[0] == DINIT_RP_CPVERSION);
+    assert(wdata.size() == 5);
+    assert(wdata[0] == (char)cp_rply::CPVERSION);
 
-	delete cc;
+    delete cc;
 }
 
 void cptest_listservices()
 {
-	service_set sset;
+    service_set sset;
 
-	service_record *s1 = new service_record(&sset, "test-service-1", service_type_t::INTERNAL, {});
-	sset.add_service(s1);
-	service_record *s2 = new service_record(&sset, "test-service-2", service_type_t::INTERNAL, {});
-	sset.add_service(s2);
-	service_record *s3 = new service_record(&sset, "test-service-3", service_type_t::INTERNAL, {});
-	sset.add_service(s3);
+    service_record *s1 = new service_record(&sset, "test-service-1", service_type_t::INTERNAL, {});
+    sset.add_service(s1);
+    service_record *s2 = new service_record(&sset, "test-service-2", service_type_t::INTERNAL, {});
+    sset.add_service(s2);
+    service_record *s3 = new service_record(&sset, "test-service-3", service_type_t::INTERNAL, {});
+    sset.add_service(s3);
 
-	int fd = bp_sys::allocfd();
-	auto *cc = new control_conn_t(event_loop, &sset, fd);
+    int fd = bp_sys::allocfd();
+    auto *cc = new control_conn_t(event_loop, &sset, fd);
 
-	bp_sys::supply_read_data(fd, { DINIT_CP_LISTSERVICES });
+    bp_sys::supply_read_data(fd, { (char)cp_cmd::LISTSERVICES });
 
-	event_loop.regd_bidi_watchers[fd]->read_ready(event_loop, fd);
+    event_loop.regd_bidi_watchers[fd]->read_ready(event_loop, fd);
 
-	// Write will process immediately, so there's no need for this:
-	//event_loop.regd_bidi_watchers[fd]->write_ready(event_loop, fd);
+    // Write will process immediately, so there's no need for this:
+    //event_loop.regd_bidi_watchers[fd]->write_ready(event_loop, fd);
 
-	// We expect, for each service:
-	// (1 byte)   DINIT_RP_SVCINFO
-	// (1 byte)   service name length
-	// (1 byte)   state
-	// (1 byte)   target state
-	// (1 byte)   flags: has console, waiting for console, start skipped
-	// (1 byte)   stop reason
+    // We expect, for each service:
+    // (1 byte)   cp_rply::SVCINFO
+    // (1 byte)   service name length
+    // (1 byte)   state
+    // (1 byte)   target state
+    // (1 byte)   flags: has console, waiting for console, start skipped
+    // (1 byte)   stop reason
     // (2 bytes)  reserved
-	// (? bytes)  exit status (int) / process id (pid_t)
-	// (N bytes)  service name
+    // (? bytes)  exit status (int) / process id (pid_t)
+    // (N bytes)  service name
 
-	std::vector<char> wdata;
-	bp_sys::extract_written_data(fd, wdata);
+    std::vector<char> wdata;
+    bp_sys::extract_written_data(fd, wdata);
 
-	std::set<std::string> names = {"test-service-1", "test-service-2", "test-service-3"};
+    std::set<std::string> names = {"test-service-1", "test-service-2", "test-service-3"};
 
-	int pos = 0;
-	for (int i = 0; i < 3; i++) {
-		assert(wdata[pos++] == DINIT_RP_SVCINFO);
-		unsigned char name_len_c = wdata[pos++];
-		pos += 6;
+    int pos = 0;
+    for (int i = 0; i < 3; i++) {
+        assert(wdata[pos++] == (char)cp_rply::SVCINFO);
+        unsigned char name_len_c = wdata[pos++];
+        pos += 6;
 
-		pos += std::max(sizeof(int), sizeof(pid_t));
+        pos += std::max(sizeof(int), sizeof(pid_t));
 
-		std::string name;
-		for (int j = 0; j < (int)name_len_c; j++) {
-			name += wdata[pos++];
-		}
+        std::string name;
+        for (int j = 0; j < (int)name_len_c; j++) {
+            name += wdata[pos++];
+        }
 
-		// Check the service name matches one from the set, and remove it:
-		auto fn = names.find(name);
-		assert (fn != names.end());
-		names.erase(fn);
-	}
+        // Check the service name matches one from the set, and remove it:
+        auto fn = names.find(name);
+        assert (fn != names.end());
+        names.erase(fn);
+    }
 
-	delete cc;
+    delete cc;
 }
 
-static control_conn_t::handle_t  find_service(int fd, const char *service_name,
+static handle_t  find_service(int fd, const char *service_name,
         service_state_t expected_state, service_state_t expected_target_state)
 {
-    std::vector<char> cmd = { DINIT_CP_FINDSERVICE };
+    std::vector<char> cmd = { (char)cp_cmd::FINDSERVICE };
     uint16_t name_len = strlen(service_name);
     char *name_len_cptr = reinterpret_cast<char *>(&name_len);
     cmd.insert(cmd.end(), name_len_cptr, name_len_cptr + sizeof(name_len));
@@ -124,19 +133,19 @@ static control_conn_t::handle_t  find_service(int fd, const char *service_name,
     bp_sys::extract_written_data(fd, wdata);
 
     // We expect:
-    // (1 byte)   DINIT_RP_SERVICERECORD
+    // (1 byte)   cp_rply::SERVICERECORD
     // (1 byte)   state
     // (handle_t) handle
     // (1 byte)   target state
 
-    assert(wdata.size() == 3 + sizeof(control_conn_t::handle_t));
-    assert(wdata[0] == DINIT_RP_SERVICERECORD);
+    assert(wdata.size() == 3 + sizeof(handle_t));
+    assert(wdata[0] == (char)cp_rply::SERVICERECORD);
     service_state_t s = static_cast<service_state_t>(wdata[1]);
     assert(s == expected_state);
     service_state_t ts = static_cast<service_state_t>(wdata[6]);
     assert(ts == expected_target_state);
 
-    control_conn_t::handle_t h1;
+    handle_t h1;
     std::copy(wdata.data() + 2, wdata.data() + 2 + sizeof(h1), reinterpret_cast<char *>(&h1));
 
     return h1;
@@ -207,7 +216,7 @@ void cptest_findservice3()
     int fd = bp_sys::allocfd();
     auto *cc = new control_conn_t(event_loop, &sset, fd);
 
-    std::vector<char> cmd = { DINIT_CP_FINDSERVICE };
+    std::vector<char> cmd = { (char)cp_cmd::FINDSERVICE };
     uint16_t name_len = strlen(service_name_2);
     char *name_len_cptr = reinterpret_cast<char *>(&name_len);
     cmd.insert(cmd.end(), name_len_cptr, name_len_cptr + sizeof(name_len));
@@ -218,13 +227,13 @@ void cptest_findservice3()
     event_loop.regd_bidi_watchers[fd]->read_ready(event_loop, fd);
 
     // We expect:
-    // (1 byte)   DINIT_RP_NOSERVICE
+    // (1 byte)   cp_rply::NOSERVICE
 
     std::vector<char> wdata;
     bp_sys::extract_written_data(fd, wdata);
 
     assert(wdata.size() == 1);
-    assert(wdata[0] == DINIT_RP_NOSERVICE);
+    assert(wdata[0] == (char)cp_rply::NOSERVICE);
 
     delete cc;
 }
@@ -265,7 +274,7 @@ void cptest_loadservice()
     int fd = bp_sys::allocfd();
     auto *cc = new control_conn_t(event_loop, &sset, fd);
 
-    std::vector<char> cmd = { DINIT_CP_LOADSERVICE };
+    std::vector<char> cmd = { (char)cp_cmd::LOADSERVICE };
     uint16_t name_len = strlen(service_name_1);
     char *name_len_cptr = reinterpret_cast<char *>(&name_len);
     cmd.insert(cmd.end(), name_len_cptr, name_len_cptr + sizeof(name_len));
@@ -277,7 +286,7 @@ void cptest_loadservice()
     event_loop.regd_bidi_watchers[fd]->read_ready(event_loop, fd);
 
     // We expect:
-    // (1 byte)   DINIT_RP_SERVICERECORD
+    // (1 byte)   cp_rply::SERVICERECORD
     // (1 byte)   state
     // (handle_t) handle
     // (1 byte)   target state
@@ -285,8 +294,8 @@ void cptest_loadservice()
     std::vector<char> wdata;
     bp_sys::extract_written_data(fd, wdata);
 
-    assert(wdata.size() == 3 + sizeof(control_conn_t::handle_t));
-    assert(wdata[0] == DINIT_RP_SERVICERECORD);
+    assert(wdata.size() == 3 + sizeof(handle_t));
+    assert(wdata[0] == (char)cp_rply::SERVICERECORD);
     service_state_t s = static_cast<service_state_t>(wdata[1]);
     assert(s == service_state_t::STOPPED);
     service_state_t ts = static_cast<service_state_t>(wdata[6]);
@@ -295,7 +304,7 @@ void cptest_loadservice()
     assert(sset.service1 != nullptr);
     assert(sset.service2 == nullptr);
 
-    cmd = { DINIT_CP_LOADSERVICE };
+    cmd = { (char)cp_cmd::LOADSERVICE };
     cmd.insert(cmd.end(), name_len_cptr, name_len_cptr + sizeof(name_len));
     cmd.insert(cmd.end(), service_name_2, service_name_2 + name_len);
 
@@ -304,15 +313,15 @@ void cptest_loadservice()
     event_loop.regd_bidi_watchers[fd]->read_ready(event_loop, fd);
 
     // We expect:
-    // (1 byte)   DINIT_RP_SERVICERECORD
+    // (1 byte)   cp_rply::SERVICERECORD
     // (1 byte)   state
     // (handle_t) handle
     // (1 byte)   target state
 
     bp_sys::extract_written_data(fd, wdata);
 
-    assert(wdata.size() == 3 + sizeof(control_conn_t::handle_t));
-    assert(wdata[0] == DINIT_RP_SERVICERECORD);
+    assert(wdata.size() == 3 + sizeof(handle_t));
+    assert(wdata[0] == (char)cp_rply::SERVICERECORD);
     s = static_cast<service_state_t>(wdata[1]);
     assert(s == service_state_t::STOPPED);
     ts = static_cast<service_state_t>(wdata[6]);
@@ -337,11 +346,11 @@ void cptest_startstop()
     auto *cc = new control_conn_t(event_loop, &sset, fd);
 
     // Get a service handle:
-    control_conn_t::handle_t h = find_service(fd, service_name, service_state_t::STOPPED,
+    handle_t h = find_service(fd, service_name, service_state_t::STOPPED,
             service_state_t::STOPPED);
 
     // Issue start:
-    std::vector<char> cmd = { DINIT_CP_STARTSERVICE, 0 /* don't pin */ };
+    std::vector<char> cmd = { (char)cp_cmd::STARTSERVICE, 0 /* don't pin */ };
     char * h_cp = reinterpret_cast<char *>(&h);
     cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h));
 
@@ -352,21 +361,21 @@ void cptest_startstop()
     std::vector<char> wdata;
     bp_sys::extract_written_data(fd, wdata);
     assert(wdata.size() == 1 + 7 + STATUS_BUFFER_SIZE /* ACK reply + info packet */);
-    assert(wdata[0] == DINIT_IP_SERVICEEVENT);
+    assert(wdata[0] == (char)cp_info::SERVICEEVENT);
     // packetsize, key (handle), event
     assert(wdata[1] == 7 + STATUS_BUFFER_SIZE);
-    control_conn_t::handle_t ip_h;
+    handle_t ip_h;
     std::copy(wdata.data() + 2, wdata.data() + 2 + sizeof(ip_h), reinterpret_cast<char *>(&ip_h));
     assert(ip_h == h);
     assert(wdata[6] == static_cast<int>(service_event_t::STARTED));
 
     constexpr unsigned reply_start = 7 + STATUS_BUFFER_SIZE;
     // we get ALREADYSS since it started immediately:
-    assert(wdata[reply_start] == DINIT_RP_ALREADYSS);
+    assert(wdata[reply_start] == (char)cp_rply::ALREADYSS);
     assert(s1->get_state() == service_state_t::STARTED);
 
     // Issue stop:
-    cmd = { DINIT_CP_STOPSERVICE, 0 /* don't pin */ };
+    cmd = { (char)cp_cmd::STOPSERVICE, 0 /* don't pin */ };
     cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h));
 
     bp_sys::supply_read_data(fd, std::move(cmd));
@@ -375,14 +384,14 @@ void cptest_startstop()
 
     bp_sys::extract_written_data(fd, wdata);
     assert(wdata.size() == 1 + 7 + STATUS_BUFFER_SIZE);
-    assert(wdata[0] == DINIT_IP_SERVICEEVENT);
+    assert(wdata[0] == (char)cp_info::SERVICEEVENT);
     // packetsize, key (handle), event
     assert(wdata[1] == 7 + STATUS_BUFFER_SIZE);
     std::copy(wdata.data() + 2, wdata.data() + 2 + sizeof(ip_h), reinterpret_cast<char *>(&ip_h));
     assert(ip_h == h);
     assert(wdata[6] == static_cast<int>(service_event_t::STOPPED));
     // we get ALREADYSS since it stopped immediately:
-    assert(wdata[reply_start] == DINIT_RP_ALREADYSS);
+    assert(wdata[reply_start] == (char)cp_rply::ALREADYSS);
     assert(s1->get_state() == service_state_t::STOPPED);
 
     delete cc;
@@ -403,11 +412,11 @@ void cptest_start_pinned()
     s1->pin_stop();
 
     // Get a service handle:
-    control_conn_t::handle_t h = find_service(fd, service_name, service_state_t::STOPPED,
+    handle_t h = find_service(fd, service_name, service_state_t::STOPPED,
             service_state_t::STOPPED);
 
     // Issue start:
-    std::vector<char> cmd = { DINIT_CP_STARTSERVICE, 0 /* don't pin */ };
+    std::vector<char> cmd = { (char)cp_cmd::STARTSERVICE, 0 /* don't pin */ };
     char * h_cp = reinterpret_cast<char *>(&h);
     cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h));
 
@@ -417,8 +426,8 @@ void cptest_start_pinned()
 
     std::vector<char> wdata;
     bp_sys::extract_written_data(fd, wdata);
-    assert(wdata.size() == 1 /* DINIT_RP_PINNEDSTOPPED */);
-    assert(wdata[0] == DINIT_RP_PINNEDSTOPPED);
+    assert(wdata.size() == 1 /* cp_rply::PINNEDSTOPPED */);
+    assert(wdata[0] == (char)cp_rply::PINNEDSTOPPED);
 
     delete cc;
 }
@@ -444,13 +453,13 @@ void cptest_gentlestop()
     int fd = bp_sys::allocfd();
     auto *cc = new control_conn_t(event_loop, &sset, fd);
 
-    control_conn_t::handle_t h = find_service(fd, test1_name, service_state_t::STARTED,
+    handle_t h = find_service(fd, test1_name, service_state_t::STARTED,
             service_state_t::STARTED);
 
     char * h_cp = reinterpret_cast<char *>(&h);
 
     // Issue stop:
-    std::vector<char> cmd = { DINIT_CP_STOPSERVICE, 2 /* don't pin, gentle */ };
+    std::vector<char> cmd = { (char)cp_cmd::STOPSERVICE, 2 /* don't pin, gentle */ };
     cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h));
 
     bp_sys::supply_read_data(fd, std::move(cmd));
@@ -461,18 +470,18 @@ void cptest_gentlestop()
     bp_sys::extract_written_data(fd, wdata);
 
     // We expect:
-    // 1 byte: DINIT_RP_DEPENDENTS
+    // 1 byte: cp_rply::DEPENDENTS
     // size_t: number of handles (N)
     // N * handle_t: handles for dependents that would be stopped
 
-    assert(wdata.size() == (1 + sizeof(size_t) + sizeof(control_conn_t::handle_t)));
-    assert(wdata[0] == DINIT_RP_DEPENDENTS);
+    assert(wdata.size() == (1 + sizeof(size_t) + sizeof(handle_t)));
+    assert(wdata[0] == (char)cp_rply::DEPENDENTS);
 
     size_t nhandles;
     memcpy(&nhandles, wdata.data() + 1, sizeof(nhandles));
     assert(nhandles == 1);
 
-    control_conn_t::handle_t rhandle;
+    handle_t rhandle;
     memcpy(&rhandle, wdata.data() + 1 + sizeof(size_t), sizeof(rhandle));
 
     service_record * rservice = control_conn_t_test::service_from_handle(cc, rhandle);
@@ -494,13 +503,13 @@ void cptest_queryname()
     auto *cc = new control_conn_t(event_loop, &sset, fd);
 
     // Get a service handle:
-    control_conn_t::handle_t h = find_service(fd, test1_name, service_state_t::STOPPED,
+    handle_t h = find_service(fd, test1_name, service_state_t::STOPPED,
             service_state_t::STOPPED);
 
     char * h_cp = reinterpret_cast<char *>(&h);
 
     // Issue name query:
-    std::vector<char> cmd = { DINIT_CP_QUERYSERVICENAME, 0 /* reserved */ };
+    std::vector<char> cmd = { (char)cp_cmd::QUERYSERVICENAME, 0 /* reserved */ };
     cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h));
 
     bp_sys::supply_read_data(fd, std::move(cmd));
@@ -511,13 +520,13 @@ void cptest_queryname()
     bp_sys::extract_written_data(fd, wdata);
 
     // We expect:
-    // 1 byte packet type = DINIT_RP_SERVICENAME
+    // 1 byte packet type = cp_rply::SERVICENAME
     // 1 byte reserved
     // uint16_t length
     // N bytes name
 
     assert(wdata.size() == (2 + sizeof(uint16_t) + strlen(test1_name)));
-    assert(wdata[0] == DINIT_RP_SERVICENAME);
+    assert(wdata[0] == (char)cp_rply::SERVICENAME);
     assert(wdata[1] == 0);
     uint16_t len;
     memcpy(&len, wdata.data() + 2, sizeof(uint16_t));
@@ -544,11 +553,11 @@ void cptest_unload()
     int fd = bp_sys::allocfd();
     auto *cc = new control_conn_t(event_loop, &sset, fd);
 
-    control_conn_t::handle_t h1 = find_service(fd, service_name1, service_state_t::STOPPED,
+    handle_t h1 = find_service(fd, service_name1, service_state_t::STOPPED,
             service_state_t::STOPPED);
 
     // Issue unload:
-    std::vector<char> cmd = { DINIT_CP_UNLOADSERVICE };
+    std::vector<char> cmd = { (char)cp_cmd::UNLOADSERVICE };
     char * h_cp = reinterpret_cast<char *>(&h1);
     cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h1));
 
@@ -560,15 +569,15 @@ void cptest_unload()
     std::vector<char> wdata;
     bp_sys::extract_written_data(fd, wdata);
     assert(wdata.size() == 1);
-    assert(wdata[0] == DINIT_RP_NAK);
+    assert(wdata[0] == (char)cp_rply::NAK);
 
 
-    control_conn_t::handle_t h2 = find_service(fd, service_name2, service_state_t::STOPPED,
+    handle_t h2 = find_service(fd, service_name2, service_state_t::STOPPED,
             service_state_t::STOPPED);
 
     // Issue unload for s2:
 
-    cmd = { DINIT_CP_UNLOADSERVICE };
+    cmd = { (char)cp_cmd::UNLOADSERVICE };
     h_cp = reinterpret_cast<char *>(&h2);
     cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h2));
 
@@ -579,11 +588,11 @@ void cptest_unload()
     // We should receive ACK:
     bp_sys::extract_written_data(fd, wdata);
     assert(wdata.size() == 1);
-    assert(wdata[0] == DINIT_RP_ACK);
+    assert(wdata[0] == (char)cp_rply::ACK);
 
     // Now try to unload s1 again:
 
-    cmd = { DINIT_CP_UNLOADSERVICE };
+    cmd = { (char)cp_cmd::UNLOADSERVICE };
     h_cp = reinterpret_cast<char *>(&h1);
     cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h1));
 
@@ -594,10 +603,10 @@ void cptest_unload()
     // We should receive ACK:
     bp_sys::extract_written_data(fd, wdata);
     assert(wdata.size() == 1);
-    assert(wdata[0] == DINIT_RP_ACK);
+    assert(wdata[0] == (char)cp_rply::ACK);
 
     // If we try to FIND service 1 now, it should not be there:
-    cmd = { DINIT_CP_FINDSERVICE };
+    cmd = { (char)cp_cmd::FINDSERVICE };
     uint16_t name_len = strlen(service_name1);
     char *name_len_cptr = reinterpret_cast<char *>(&name_len);
     cmd.insert(cmd.end(), name_len_cptr, name_len_cptr + sizeof(name_len));
@@ -609,7 +618,7 @@ void cptest_unload()
 
     bp_sys::extract_written_data(fd, wdata);
     assert(wdata.size() == 1);
-    assert(wdata[0] == DINIT_RP_NOSERVICE);
+    assert(wdata[0] == (char)cp_rply::NOSERVICE);
 
     delete cc;
 }
@@ -629,13 +638,13 @@ void cptest_addrmdeps()
     int fd = bp_sys::allocfd();
     auto *cc = new control_conn_t(event_loop, &sset, fd);
 
-    control_conn_t::handle_t h1 = find_service(fd, service_name1, service_state_t::STOPPED,
+    handle_t h1 = find_service(fd, service_name1, service_state_t::STOPPED,
             service_state_t::STOPPED);
-    control_conn_t::handle_t h2 = find_service(fd, service_name2, service_state_t::STOPPED,
+    handle_t h2 = find_service(fd, service_name2, service_state_t::STOPPED,
             service_state_t::STOPPED);
 
     // Add dep from s1 -> s2:
-    std::vector<char> cmd = { DINIT_CP_ADD_DEP, static_cast<char>(dependency_type::REGULAR) };
+    std::vector<char> cmd = { (char)cp_cmd::ADD_DEP, static_cast<char>(dependency_type::REGULAR) };
     char * h1cp = reinterpret_cast<char *>(&h1);
     char * h2cp = reinterpret_cast<char *>(&h2);
     cmd.insert(cmd.end(), h1cp, h1cp + sizeof(h1));
@@ -647,10 +656,10 @@ void cptest_addrmdeps()
     bp_sys::extract_written_data(fd, wdata);
 
     assert(wdata.size() == 1);
-    assert(wdata[0] == DINIT_RP_ACK);
+    assert(wdata[0] == (char)cp_rply::ACK);
 
     // Issue start for S1. S2 should also start:
-    cmd = { DINIT_CP_STARTSERVICE, 0 /* don't pin */ };
+    cmd = { (char)cp_cmd::STARTSERVICE, 0 /* don't pin */ };
     cmd.insert(cmd.end(), h1cp, h1cp + sizeof(h1));
 
     bp_sys::supply_read_data(fd, std::move(cmd));
@@ -662,7 +671,7 @@ void cptest_addrmdeps()
     assert(s2->get_state() == service_state_t::STARTED);
 
     // Remove dependency from S1 -> S2:
-    cmd = { DINIT_CP_REM_DEP, static_cast<char>(dependency_type::REGULAR) };
+    cmd = { (char)cp_cmd::REM_DEP, static_cast<char>(dependency_type::REGULAR) };
     cmd.insert(cmd.end(), h1cp, h1cp + sizeof(h1));
     cmd.insert(cmd.end(), h2cp, h2cp + sizeof(h2));
 
@@ -694,11 +703,11 @@ void cptest_enableservice()
     int fd = bp_sys::allocfd();
     auto *cc = new control_conn_t(event_loop, &sset, fd);
 
-    control_conn_t::handle_t h1 = find_service(fd, service_name1, service_state_t::STARTED, service_state_t::STARTED);
-    control_conn_t::handle_t h2 = find_service(fd, service_name2, service_state_t::STOPPED, service_state_t::STOPPED);
+    handle_t h1 = find_service(fd, service_name1, service_state_t::STARTED, service_state_t::STARTED);
+    handle_t h2 = find_service(fd, service_name2, service_state_t::STOPPED, service_state_t::STOPPED);
 
     // Enable from s1 -> s2:
-    std::vector<char> cmd = { DINIT_CP_ENABLESERVICE, static_cast<char>(dependency_type::WAITS_FOR) };
+    std::vector<char> cmd = { (char)cp_cmd::ENABLESERVICE, static_cast<char>(dependency_type::WAITS_FOR) };
     char * h1cp = reinterpret_cast<char *>(&h1);
     char * h2cp = reinterpret_cast<char *>(&h2);
     cmd.insert(cmd.end(), h1cp, h1cp + sizeof(h1));
@@ -711,16 +720,16 @@ void cptest_enableservice()
     bp_sys::extract_written_data(fd, wdata);
 
     assert(wdata.size() == 1 + 7 + STATUS_BUFFER_SIZE /* ACK reply + info packet */);
-    assert(wdata[0] == DINIT_IP_SERVICEEVENT);
+    assert(wdata[0] == (char)cp_info::SERVICEEVENT);
     // packetsize, key (handle), event
     assert(wdata[1] == 7 + STATUS_BUFFER_SIZE);
-    control_conn_t::handle_t ip_h;
+    handle_t ip_h;
     std::copy(wdata.data() + 2, wdata.data() + 2 + sizeof(ip_h), reinterpret_cast<char *>(&ip_h));
     assert(ip_h == h2);
     assert(wdata[6] == static_cast<int>(service_event_t::STARTED));
 
     // and then the ack:
-    assert(wdata[7 + STATUS_BUFFER_SIZE] == DINIT_RP_ACK);
+    assert(wdata[7 + STATUS_BUFFER_SIZE] == (char)cp_rply::ACK);
 
     sset.process_queues();
 
@@ -750,7 +759,7 @@ void cptest_restart()
     auto *cc = new control_conn_t(event_loop, &sset, fd);
 
     // Get a service handle:
-    control_conn_t::handle_t h = find_service(fd, service_name, service_state_t::STOPPED,
+    handle_t h = find_service(fd, service_name, service_state_t::STOPPED,
             service_state_t::STOPPED);
 
     std::vector<char> wdata;
@@ -758,7 +767,7 @@ void cptest_restart()
     assert(wdata.size() == 0);
 
     // Issue restart:
-    std::vector<char> cmd = { DINIT_CP_STOPSERVICE, 4 /* restart */ };
+    std::vector<char> cmd = { (char)cp_cmd::STOPSERVICE, 4 /* restart */ };
     char * h_cp = reinterpret_cast<char *>(&h);
     cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h));
 
@@ -769,7 +778,7 @@ void cptest_restart()
     bp_sys::extract_written_data(fd, wdata);
 
     assert(wdata.size() == 1 /* NAK reply, wrong state */);
-    assert(wdata[0] == DINIT_RP_NAK);
+    assert(wdata[0] == (char)cp_rply::NAK);
 
     // Start the service now:
     s1->start();
@@ -785,13 +794,13 @@ void cptest_restart()
     bp_sys::extract_written_data(fd, wdata);
 
     assert(wdata.size() == 7 + STATUS_BUFFER_SIZE + 1);  // info packet (service stopped) + ACK
-    assert(wdata[0] == DINIT_IP_SERVICEEVENT);
+    assert(wdata[0] == (char)cp_info::SERVICEEVENT);
     assert(wdata[1] == 7 + STATUS_BUFFER_SIZE);
-    control_conn_t::handle_t ip_h;
+    handle_t ip_h;
     std::copy(wdata.data() + 2, wdata.data() + 2 + sizeof(ip_h), reinterpret_cast<char *>(&ip_h));
     assert(ip_h == h);
     assert(wdata[6] == static_cast<int>(service_event_t::STOPPED));
-    assert(wdata[7 + STATUS_BUFFER_SIZE] == DINIT_RP_ACK);
+    assert(wdata[7 + STATUS_BUFFER_SIZE] == (char)cp_rply::ACK);
 
     sset.process_queues();
     assert(s1->get_state() == service_state_t::STARTING);
@@ -803,7 +812,7 @@ void cptest_restart()
     bp_sys::extract_written_data(fd, wdata);
 
     assert(wdata.size() == 7 + STATUS_BUFFER_SIZE);  /* info packet */
-    assert(wdata[0] == DINIT_IP_SERVICEEVENT);
+    assert(wdata[0] == (char)cp_info::SERVICEEVENT);
     // packetsize, key (handle), event
     assert(wdata[1] == 7 + STATUS_BUFFER_SIZE);
     std::copy(wdata.data() + 2, wdata.data() + 2 + sizeof(ip_h), reinterpret_cast<char *>(&ip_h));
@@ -838,11 +847,11 @@ void cptest_wake()
     int fd = bp_sys::allocfd();
     auto *cc = new control_conn_t(event_loop, &sset, fd);
 
-    control_conn_t::handle_t h1 = find_service(fd, service_name1, service_state_t::STOPPED,
+    handle_t h1 = find_service(fd, service_name1, service_state_t::STOPPED,
             service_state_t::STOPPED);
 
     // Wake s1:
-    std::vector<char> cmd = { DINIT_CP_WAKESERVICE, 0 /* don't pin */ };
+    std::vector<char> cmd = { (char)cp_cmd::WAKESERVICE, 0 /* don't pin */ };
     char * h_cp = reinterpret_cast<char *>(&h1);
     cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h1));
     bp_sys::supply_read_data(fd, std::move(cmd));
@@ -852,16 +861,16 @@ void cptest_wake()
     bp_sys::extract_written_data(fd, wdata);
 
     assert(wdata.size() == 1 + 7 + STATUS_BUFFER_SIZE /* ACK reply + info packet */);
-    assert(wdata[0] == DINIT_IP_SERVICEEVENT);
+    assert(wdata[0] == (char)cp_info::SERVICEEVENT);
     // packetsize, key (handle), event
     assert(wdata[1] == 7 + STATUS_BUFFER_SIZE);
-    control_conn_t::handle_t ip_h;
+    handle_t ip_h;
     std::copy(wdata.data() + 2, wdata.data() + 2 + sizeof(ip_h), reinterpret_cast<char *>(&ip_h));
     assert(ip_h == h1);
     assert(wdata[6] == static_cast<int>(service_event_t::STARTED));
 
     // and then the ack (already started):
-    assert(wdata[7 + STATUS_BUFFER_SIZE] == DINIT_RP_ALREADYSS);
+    assert(wdata[7 + STATUS_BUFFER_SIZE] == (char)cp_rply::ALREADYSS);
 
     // now stop s2 (and therefore s1):
     s2->stop(true);
@@ -873,7 +882,7 @@ void cptest_wake()
     bp_sys::extract_written_data(fd, wdata);
 
     // Trying to wake s1 should now fail:
-    cmd = { DINIT_CP_WAKESERVICE, 0 /* don't pin */ };
+    cmd = { (char)cp_cmd::WAKESERVICE, 0 /* don't pin */ };
     cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h1));
     bp_sys::supply_read_data(fd, std::move(cmd));
 
@@ -881,7 +890,7 @@ void cptest_wake()
     bp_sys::extract_written_data(fd, wdata);
 
     assert(wdata.size() == 1);
-    assert(wdata[0] == DINIT_RP_NAK);
+    assert(wdata[0] == (char)cp_rply::NAK);
 
     delete cc;
 }
@@ -905,11 +914,11 @@ void cptest_servicestatus()
 
     auto STOPPED = service_state_t::STOPPED;
     auto STARTED = service_state_t::STARTED;
-    control_conn_t::handle_t h1 = find_service(fd, "test-service-1", STOPPED, STOPPED);
-    control_conn_t::handle_t h2 = find_service(fd, "test-service-2", STARTED, STARTED);
-    control_conn_t::handle_t h3 = find_service(fd, "test-service-3", STOPPED, STOPPED);
+    handle_t h1 = find_service(fd, "test-service-1", STOPPED, STOPPED);
+    handle_t h2 = find_service(fd, "test-service-2", STARTED, STARTED);
+    handle_t h3 = find_service(fd, "test-service-3", STOPPED, STOPPED);
 
-    std::vector<char> cmd = { DINIT_CP_SERVICESTATUS };
+    std::vector<char> cmd = { (char)cp_cmd::SERVICESTATUS };
     char * h_cp = reinterpret_cast<char *>(&h1);
     cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h1));
     bp_sys::supply_read_data(fd, std::move(cmd));
@@ -921,17 +930,17 @@ void cptest_servicestatus()
 
     constexpr static int STATUS_BUFFER_SIZE = 6 + ((sizeof(pid_t) > sizeof(int)) ? sizeof(pid_t) : sizeof(int));
 
-    // 1 byte: DINIT_RP_SERVICESTATUS
+    // 1 byte: cp_rply::SERVICESTATUS
     // 1 byte: reserved
     // STATUS_BUFFER_SIZE bytes: status
     assert(wdata.size() == (2 + STATUS_BUFFER_SIZE));
-    assert(wdata[0] == DINIT_RP_SERVICESTATUS);
+    assert(wdata[0] == (char)cp_rply::SERVICESTATUS);
     assert(wdata[2] == (int)service_state_t::STOPPED); // state
     assert(wdata[3] == (int)service_state_t::STOPPED); // target state
     assert(wdata[4] == 0); // various flags
 
     cmd.clear();
-    cmd.push_back(DINIT_CP_SERVICESTATUS);
+    cmd.push_back((char)cp_cmd::SERVICESTATUS);
     h_cp = reinterpret_cast<char *>(&h2);
     cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h1));
     bp_sys::supply_read_data(fd, std::move(cmd));
@@ -941,7 +950,7 @@ void cptest_servicestatus()
     bp_sys::extract_written_data(fd, wdata);
 
     assert(wdata.size() == (2 + STATUS_BUFFER_SIZE));
-    assert(wdata[0] == DINIT_RP_SERVICESTATUS);
+    assert(wdata[0] == (char)cp_rply::SERVICESTATUS);
     assert(wdata[2] == (int)service_state_t::STARTED); // state
     assert(wdata[3] == (int)service_state_t::STARTED); // target state
     assert(wdata[4] == 8); // various flags; 8 = marked active
@@ -951,6 +960,155 @@ void cptest_servicestatus()
     delete cc;
 }
 
+void cptest_sendsignal()
+{
+    using namespace std;
+
+    service_set sset;
+    ha_string command = "test-command";
+    list<pair<unsigned,unsigned>> command_offsets;
+    command_offsets.emplace_back(0, command.length());
+    std::list<prelim_dep> depends;
+
+    process_service p {&sset, "test-service", std::move(command), command_offsets, depends};
+    init_service_defaults(p);
+    sset.add_service(&p);
+
+    p.start();
+    sset.process_queues();
+    base_process_service_test::exec_succeeded(&p);
+    sset.process_queues();
+
+    int fd = bp_sys::allocfd();
+    auto *cc = new control_conn_t(event_loop, &sset, fd);
+
+    // Get a service handle:
+    handle_t h = find_service(fd, "test-service", service_state_t::STARTED,
+            service_state_t::STARTED);
+
+    // Prepare a signal: (SIGHUP for example)
+    sig_num_t sig = SIGHUP;
+
+    // Issue a signal:
+    std::vector<char> cmd = { (char)cp_cmd::SIGNAL };
+    char * sig_cp = reinterpret_cast<char *>(&sig);
+    char * h_cp = reinterpret_cast<char *>(&h);
+    cmd.insert(cmd.end(), sig_cp, sig_cp + sizeof(sig));
+    cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h));
+
+    bp_sys::supply_read_data(fd, std::move(cmd));
+    event_loop.regd_bidi_watchers[fd]->read_ready(event_loop, fd);
+
+    std::vector<char> wdata;
+    bp_sys::extract_written_data(fd, wdata);
+    assert(wdata.size() == 1);
+    assert(wdata[0] == (char)cp_rply::ACK);
+
+    assert(bp_sys::last_sig_sent == SIGHUP);
+
+    // Prepare an another signal: (for sure)
+    sig = SIGILL;
+
+    // Issue a signal:
+    cmd = { (char)cp_cmd::SIGNAL };
+    sig_cp = reinterpret_cast<char *>(&sig);
+    cmd.insert(cmd.end(), sig_cp, sig_cp + sizeof(sig));
+    cmd.insert(cmd.end(), h_cp, h_cp + sizeof(h));
+
+    bp_sys::supply_read_data(fd, std::move(cmd));
+    event_loop.regd_bidi_watchers[fd]->read_ready(event_loop, fd);
+
+    wdata.clear();
+    bp_sys::extract_written_data(fd, wdata);
+    assert(wdata.size() == 1);
+    assert(wdata[0] == (char)cp_rply::ACK);
+
+    assert(bp_sys::last_sig_sent == SIGILL);
+
+    sset.remove_service(&p);
+
+    delete cc;
+}
+
+// Two commands in one packet
+void cptest_two_commands()
+{
+    service_set sset;
+
+    const char * const service_name_1 = "test-service-1";
+    const char * const service_name_2 = "test-service-2";
+    const char * const service_name_3 = "test-service-3";
+
+    service_record *s1 = new service_record(&sset, service_name_1, service_type_t::INTERNAL, {});
+    sset.add_service(s1);
+    service_record *s2 = new service_record(&sset, service_name_2, service_type_t::INTERNAL, {});
+    sset.add_service(s2);
+    service_record *s3 = new service_record(&sset, service_name_3, service_type_t::INTERNAL, {});
+    sset.add_service(s3);
+
+    int fd = bp_sys::allocfd();
+    auto *cc = new control_conn_t(event_loop, &sset, fd);
+
+    std::vector<char> cmd = { (char)cp_cmd::FINDSERVICE };
+    uint16_t name_len = strlen(service_name_1);
+    char *name_len_cptr = reinterpret_cast<char *>(&name_len);
+    cmd.insert(cmd.end(), name_len_cptr, name_len_cptr + sizeof(name_len));
+    cmd.insert(cmd.end(), service_name_1, service_name_1 + name_len);
+
+    // Insert a 2nd FINDSERVICE command into the same vector:
+    cmd.push_back((char)cp_cmd::FINDSERVICE);
+    name_len = strlen(service_name_2);
+    cmd.insert(cmd.end(), name_len_cptr, name_len_cptr + sizeof(name_len));
+    cmd.insert(cmd.end(), service_name_2, service_name_2 + name_len);
+
+    bp_sys::supply_read_data(fd, std::move(cmd));
+
+    event_loop.regd_bidi_watchers[fd]->read_ready(event_loop, fd);
+
+    std::vector<char> wdata;
+    bp_sys::extract_written_data(fd, wdata);
+
+    // We expect 2x:
+    // (1 byte)   cp_rply::SERVICERECORD
+    // (1 byte)   state
+    // (handle_t) handle
+    // (1 byte)   target state
+
+    assert(wdata.size() == 2 * (3 + sizeof(handle_t)));
+    assert(wdata[0] == (char)cp_rply::SERVICERECORD);
+    assert(wdata[3 + sizeof(handle_t)] == (char)cp_rply::SERVICERECORD);
+
+    delete cc;
+}
+
+void cptest_closehandle()
+{
+    service_set sset;
+
+    const char * const service_name_1 = "test-service-1";
+
+    service_record *s1 = new service_record(&sset, service_name_1, service_type_t::INTERNAL, {});
+    sset.add_service(s1);
+
+    int fd = bp_sys::allocfd();
+    auto *cc = new control_conn_t(event_loop, &sset, fd);
+
+    handle_t hndl = find_service(fd, service_name_1, service_state_t::STOPPED, service_state_t::STOPPED);
+
+    std::vector<char> cmd = { (char)cp_cmd::CLOSEHANDLE };
+    cmd.insert(cmd.end(), (char *)&hndl, (char *)&hndl + sizeof(hndl));
+    bp_sys::supply_read_data(fd, std::move(cmd));
+    event_loop.regd_bidi_watchers[fd]->read_ready(event_loop, fd);
+
+    std::vector<char> wdata;
+    bp_sys::extract_written_data(fd, wdata);
+
+    assert(wdata.size() == 1);
+    assert(wdata[0] == (char)cp_rply::ACK);
+
+    delete cc;
+}
+
 
 #define RUN_TEST(name, spacing) \
     std::cout << #name "..." spacing << std::flush; \
@@ -975,5 +1133,8 @@ int main(int argc, char **argv)
     RUN_TEST(cptest_restart, "            ");
     RUN_TEST(cptest_wake, "               ");
     RUN_TEST(cptest_servicestatus, "      ");
+    RUN_TEST(cptest_sendsignal, "         ");
+    RUN_TEST(cptest_two_commands, "       ");
+    RUN_TEST(cptest_closehandle, "        ");
     return 0;
 }
