@@ -8,78 +8,22 @@
 #include "service.h"
 #include "proc-service.h"
 #include "dinit-util.h"
+#include "test_procservice.h"
 
 // Tests of process-service related functionality.
 //
 // These tests work mostly by completely mocking out the base_process_service class. The mock
 // implementations can be found in test-baseproc.cc.
 
+#ifdef NDEBUG
+#error "This file must be built with assertions ENABLED!"
+#endif
+
 extern eventloop_t event_loop;
 
 constexpr static auto REG = dependency_type::REGULAR;
 constexpr static auto WAITS = dependency_type::WAITS_FOR;
 
-// Friend interface to access base_process_service private/protected members.
-class base_process_service_test
-{
-    public:
-    static void exec_succeeded(base_process_service *bsp)
-    {
-        bsp->waiting_for_execstat = false;
-        bsp->exec_succeeded();
-    }
-
-    static void exec_failed(base_process_service *bsp, int errcode)
-    {
-        run_proc_err err;
-        err.stage = exec_stage::DO_EXEC;
-        err.st_errno = errcode;
-    	bsp->waiting_for_execstat = false;
-    	bsp->pid = -1;
-    	bsp->exec_failed(err);
-    }
-
-    static void handle_exit(base_process_service *bsp, int exit_status)
-    {
-        bsp->pid = -1;
-        bsp->handle_exit_status(bp_sys::exit_status(true, false, exit_status));
-    }
-
-    static void handle_signal_exit(base_process_service *bsp, int signo)
-    {
-        bsp->pid = -1;
-        bsp->handle_exit_status(bp_sys::exit_status(false, true, signo));
-    }
-
-    static void handle_stop_exit(process_service *ps, int exit_status)
-    {
-        ps->stop_pid = -1;
-        ps->waiting_for_execstat = false;
-        ps->stop_status = bp_sys::exit_status(true, false, exit_status);
-        ps->handle_stop_exit();
-    }
-
-    static int get_notification_fd(base_process_service *bsp)
-    {
-        return bsp->notification_fd;
-    }
-};
-
-namespace bp_sys {
-    // last signal sent:
-    extern int last_sig_sent;
-    extern pid_t last_forked_pid;
-}
-
-static time_val default_restart_interval = time_val(0, 200000000); // 200 milliseconds
-
-static void init_service_defaults(base_process_service &ps)
-{
-    ps.set_restart_interval(time_val(10,0), 3);
-    ps.set_restart_delay(default_restart_interval); // 200 milliseconds
-    ps.set_stop_timeout(time_val(10,0));
-}
-
 // Regular service start
 void test_proc_service_start()
 {
@@ -514,6 +458,75 @@ void test_proc_term_restart4()
     sset.remove_service(&p);
 }
 
+// Failure to restart should propagate to dependent
+void test_proc_term_restart_fail()
+{
+    using namespace std;
+
+    service_set sset;
+
+    ha_string command = "test-command";
+    list<pair<unsigned,unsigned>> command_offsets;
+    command_offsets.emplace_back(0, command.length());
+    std::list<prelim_dep> depends;
+
+    process_service p {&sset, "testproc", std::move(command), command_offsets, depends};
+    init_service_defaults(p);
+    p.set_auto_restart(true);
+
+    sset.add_service(&p);
+
+    service_record p_dpt1 {&sset, "dpt1", service_type_t::INTERNAL, {{ &p, REG }}};
+    p_dpt1.set_auto_restart(true);
+    sset.add_service(&p_dpt1);
+
+    service_record p_dpt2 {&sset, "dpt2", service_type_t::INTERNAL, {{ &p_dpt1, WAITS }}};
+    sset.add_service(&p_dpt2);
+
+    p_dpt2.start();
+    sset.process_queues();
+
+    for (int i = 0; i < 3; ++i) {
+
+        assert(p.get_state() == service_state_t::STARTING);
+        base_process_service_test::exec_succeeded(&p);
+        sset.process_queues();
+
+        assert(p.get_state() == service_state_t::STARTED);
+        assert(event_loop.active_timers.size() == 0);
+
+        base_process_service_test::handle_exit(&p, 0);
+        sset.process_queues();
+
+        // Starting, restart timer should be armed:
+        assert(p.get_state() == service_state_t::STARTING);
+        assert(event_loop.active_timers.size() == 1);
+
+        event_loop.advance_time(default_restart_interval);
+
+    }
+
+    assert(p.get_state() == service_state_t::STARTING);
+    base_process_service_test::exec_succeeded(&p);
+    sset.process_queues();
+
+    assert(p.get_state() == service_state_t::STARTED);
+    assert(event_loop.active_timers.size() == 0);
+
+    // There should be no attempt to restart this time:
+
+    base_process_service_test::handle_exit(&p, 0);
+    sset.process_queues();
+
+    assert(event_loop.active_timers.size() == 0);
+    assert(p.get_state() == service_state_t::STOPPED);
+    assert(p_dpt1.get_state() == service_state_t::STOPPED);
+
+    sset.remove_service(&p_dpt2);
+    sset.remove_service(&p_dpt1);
+    sset.remove_service(&p);
+}
+
 // Termination via stop request
 void test_term_via_stop()
 {
@@ -1029,6 +1042,7 @@ void test_proc_smooth_recovery3()
     sset.remove_service(&p);
 }
 
+// stop issued during smooth recovery (waiting for restart timer)
 void test_proc_smooth_recovery4()
 {
     using namespace std;
@@ -1065,19 +1079,21 @@ void test_proc_smooth_recovery4()
 
     assert(p.get_state() == service_state_t::STARTED);
 
-    // If we stop now, timer should be cancelled
+    // If we stop now, timer should be cancelled, no signal should be sent
+    bp_sys::last_sig_sent = -1;
     p.stop(true);
 
     sset.process_queues();
 
     assert(p.get_state() == service_state_t::STOPPED);
     assert(first_instance == bp_sys::last_forked_pid);  // no more processes launched
+    assert(bp_sys::last_sig_sent == -1);
     assert(event_loop.active_timers.size() == 0);
 
     sset.remove_service(&p);
 }
 
-// stop during smooth recovery
+// stop during smooth recovery (waiting for process startup)
 void test_proc_smooth_recovery5()
 {
     using namespace std;
@@ -1143,7 +1159,7 @@ void test_proc_smooth_recovery5()
     sset.remove_service(&p);
 }
 
-// stop during smooth recovery (while waiting on restart timer)
+// smooth recovery: timeout waiting for readiness notification
 void test_proc_smooth_recovery6()
 {
     using namespace std;
@@ -1159,6 +1175,8 @@ void test_proc_smooth_recovery6()
     init_service_defaults(p);
     p.set_smooth_recovery(true);
     p.set_restart_delay(time_val {0, 1000});
+    p.set_start_timeout(time_val {1, 0});
+    p.set_notification_fd(3);
     sset.add_service(&p);
 
     p.start();
@@ -1167,6 +1185,16 @@ void test_proc_smooth_recovery6()
     base_process_service_test::exec_succeeded(&p);
     sset.process_queues();
 
+    // readiness notification from process:
+    int nfd = base_process_service_test::get_notification_fd(&p);
+    char notifystr[] = "ok started\n";
+    std::vector<char> rnotifystr;
+    rnotifystr.insert(rnotifystr.end(), notifystr, notifystr + sizeof(notifystr));
+    bp_sys::supply_read_data(nfd, std::move(rnotifystr));
+    event_loop.regd_fd_watchers[nfd]->fd_event(event_loop, nfd, dasynq::IN_EVENTS);
+    assert(p.get_state() == service_state_t::STARTED);
+    assert(event_loop.active_timers.size() == 0);
+
     pid_t first_instance = bp_sys::last_forked_pid;
 
     assert(p.get_state() == service_state_t::STARTED);
@@ -1174,20 +1202,91 @@ void test_proc_smooth_recovery6()
     base_process_service_test::handle_exit(&p, 0);
     sset.process_queues();
 
-    // since time hasn't been changed, we expect that the process has not yet been re-launched:
-    assert(first_instance == bp_sys::last_forked_pid);
+    event_loop.advance_time(time_val {0, 1000});
+
+    // new process should've been forked:
+    assert(first_instance != bp_sys::last_forked_pid);
     assert(p.get_state() == service_state_t::STARTED);
+    assert(p.get_pid() == bp_sys::last_forked_pid);
 
+    base_process_service_test::exec_succeeded(&p);
+
+    // Now, timeout while waiting for readiness:
     bp_sys::last_sig_sent = -1;
+    event_loop.advance_time(time_val {1, 0});
 
-    // Now issue a stop:
-    p.stop(true);
+    // We should see the process has been signalled:
+    assert(bp_sys::last_sig_sent == SIGINT);
+    assert(p.get_state() == service_state_t::STOPPING);
+
+    base_process_service_test::handle_exit(&p, 1);
     sset.process_queues();
 
-    // since we were waiting on the restart timer, there should be no process signalled and the
-    // state should now be stopped:
+    // The state should now be stopped:
+    assert(p.get_state() == service_state_t::STOPPED);
+    assert(event_loop.active_timers.size() == 0);
+
+    sset.remove_service(&p);
+}
+
+// smooth recovery: termination while waiting for readiness notification
+void test_proc_smooth_recovery6a()
+{
+    using namespace std;
+
+    service_set sset;
+
+    ha_string command = "test-command";
+    list<pair<unsigned,unsigned>> command_offsets;
+    command_offsets.emplace_back(0, command.length());
+    std::list<prelim_dep> depends;
+
+    process_service p {&sset, "testproc", std::move(command), command_offsets, depends};
+    init_service_defaults(p);
+    p.set_smooth_recovery(true);
+    p.set_restart_delay(time_val {0, 1000});
+    p.set_start_timeout(time_val {1, 0});
+    p.set_notification_fd(3);
+    sset.add_service(&p);
+
+    p.start();
+    sset.process_queues();
+
+    base_process_service_test::exec_succeeded(&p);
+    sset.process_queues();
+
+    // readiness notification from process:
+    int nfd = base_process_service_test::get_notification_fd(&p);
+    char notifystr[] = "ok started\n";
+    std::vector<char> rnotifystr;
+    rnotifystr.insert(rnotifystr.end(), notifystr, notifystr + sizeof(notifystr));
+    bp_sys::supply_read_data(nfd, std::move(rnotifystr));
+    event_loop.regd_fd_watchers[nfd]->fd_event(event_loop, nfd, dasynq::IN_EVENTS);
+    assert(p.get_state() == service_state_t::STARTED);
+    assert(event_loop.active_timers.size() == 0);
+
+    pid_t first_instance = bp_sys::last_forked_pid;
+
+    assert(p.get_state() == service_state_t::STARTED);
+
+    base_process_service_test::handle_exit(&p, 0);
+    sset.process_queues();
+
+    event_loop.advance_time(time_val {0, 1000});
+
+    // new process should've been forked:
+    assert(first_instance != bp_sys::last_forked_pid);
+    assert(p.get_state() == service_state_t::STARTED);
+    assert(p.get_pid() == bp_sys::last_forked_pid);
+
+    base_process_service_test::exec_succeeded(&p);
+    sset.process_queues();
+
+    base_process_service_test::handle_exit(&p, 1);
+    sset.process_queues();
+
+    // The state should now be stopped:
     assert(p.get_state() == service_state_t::STOPPED);
-    assert(bp_sys::last_sig_sent == -1);
     assert(event_loop.active_timers.size() == 0);
 
     sset.remove_service(&p);
@@ -2393,6 +2492,7 @@ int main(int argc, char **argv)
     RUN_TEST(test_proc_term_restart2, "    ");
     RUN_TEST(test_proc_term_restart3, "    ");
     RUN_TEST(test_proc_term_restart4, "    ");
+    RUN_TEST(test_proc_term_restart_fail, "");
     RUN_TEST(test_term_via_stop, "         ");
     RUN_TEST(test_term_via_stop2, "        ");
     RUN_TEST(test_term_via_stop3, "        ");
@@ -2407,6 +2507,7 @@ int main(int argc, char **argv)
     RUN_TEST(test_proc_smooth_recovery4, " ");
     RUN_TEST(test_proc_smooth_recovery5, " ");
     RUN_TEST(test_proc_smooth_recovery6, " ");
+    RUN_TEST(test_proc_smooth_recovery6a, "");
     RUN_TEST(test_bgproc_start, "          ");
     RUN_TEST(test_bgproc_start_fail, "     ");
     RUN_TEST(test_bgproc_start_fail_pid, " ");
